from __future__ import division, print_function, absolute_import
import numpy as np
from .a1sB_monomer import a1sB_eval, da1sB_dxhi00_eval, d2a1sB_dxhi00_eval
from .a1sB_monomer import d3a1sB_dxhi00_eval
from .a1sB_monomer import da1sB_dx_eval, da1sB_dx_dxhi00_eval
from .a1sB_monomer import da1sB_dx_dxhi00_dxxhi_eval
from .a1sB_monomer import da1sB_dx_d2xhi00_dxxhi_eval

from .ahs_monomer import ahs, dahs_dxhi00, d2ahs_dxhi00
from .ahs_monomer import dahs_dx, dahs_dxxhi
from .a3_monomer import a3, da3_dxhi00, d2a3_dxhi00
from .a3_monomer import da3_dx, da3_dxxhi
from .a2_monomer import a2, da2_dxhi00, d2a2_dxhi00
from .a2_monomer import da2_dx, da2_dxxhi
from .a1_monomer import a1, da1_dxhi00, d2a1_dxhi00
from .a1_monomer import da1_dx, da1_dxxhi

from .a2new_chain import da2new_dxhi00, d2a2new_dxhi00, d3a2new_dxhi00
from .a2new_chain import da2new_dx_dxhi00, da2new_dxxhi_dxhi00
from .gdHS_chain import gdHS, dgdHS_dxhi00, d2gdHS_dxhi00
from .gdHS_chain import dgdHS_dx, dgdHS_dxxhi
from .gammac_chain import gammac, dgammac_dxhi00, d2gammac_dxhi00
from .gammac_chain import dgammac_dx, dgammac_dxxhi
from .g1sigma_chain import g1sigma, dg1sigma_dxhi00, d2g1sigma_dxhi00
from .g1sigma_chain import dg1sigma_dx, dg1sigma_dxxhi
from .g2mca_chain import g2mca, dg2mca_dxhi00, d2g2mca_dxhi00
from .g2mca_chain import dg2mca_dx, dg2mca_dxxhi
from .lngmie_chain import lngmie, dlngmie_dxhi00, d2lngmie_dxhi00
from .lngmie_chain import dlngmie_dx, dlngmie_dxxhi

from .monomer_aux import dkHS_dxhi00, d2kHS_dxhi00, d3kHS_dxhi00
from .monomer_aux import dkHS_dx_dxhi00, d2kHS_dx_dxhi00

from .association_aux import Xass_solver, CIJ_matrix
from .association_aux import dXass_drho, d2Xass_drho, dXass_dx
from .association_aux import Iab, dIab_drho, d2Iab_drho, dIab


# Eq. (14) Paper 2014
def xhi_eval(xhi00, xs_k, xs_m, d_kk03):
    dxhi_dxhi00 = xs_m * np.matmul(xs_k, d_kk03)
    xhi = xhi00 * dxhi_dxhi00
    return xhi, dxhi_dxhi00


def dxhi_dx_eval(xhi00, xs_k, xs_m, d_kk03, dxk_dx_aux):
    dxhi_dxhi00 = xs_m * np.matmul(xs_k, d_kk03)
    xhi = xhi00 * dxhi_dxhi00
    dxhi_dx = (xhi00*dxk_dx_aux@d_kk03).T
    return xhi, dxhi_dxhi00, dxhi_dx


# Eq (22) Paper 2014
def xhix_eval(xhi00, xs_k, xs_m, d_kl3):
    sum1 = np.matmul(np.matmul(xs_k, d_kl3), xs_k)
    dxhix_dxhi00 = xs_m * sum1
    xhix = xhi00 * dxhix_dxhi00
    return xhix, dxhix_dxhi00


def dxhix_dx_eval(xhi00, xs_k, dxsk_dx, xs_m, zs_m, d_kl3):
    aux1 = xs_k * d_kl3
    aux2 = np.dot(xs_k, aux1)
    aux3 = aux2.sum()
    dxhix_dxhi00 = xs_m * aux3
    xhix = xhi00 * dxhix_dxhi00
    suma1 = 2*np.sum(dxsk_dx@aux1, axis=1)
    dxhix_dx_dxhi00 = (zs_m * aux3 + xs_m * suma1)
    dxhix_dx = xhi00 * dxhix_dx_dxhi00
    return xhix, dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00


def ares(self, x, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, d_kk, d_kl, d_kl3, d_kk03, x0_kl = temp_aux[:8]
    a1vdw_ctekl, a1vdwkl, x0_a1, x0_a2, I_lambdaskl = temp_aux[8:13]
    J_lambdaskl, d_ii, d_ii3, x0_ii, a1vdw_cteii, a1vdwii = temp_aux[13:19]
    tetha, x0_a1ii, x0_a2ii, x0_g1ii, x0_g2ii, I_lambdasii = temp_aux[19:25]
    J_lambdasii, x0i_matrix, beps_ii, beps_ii2 = temp_aux[25:29]

    dxhi00_drho = self.dxhi00_drho
    xhi00 = rho*dxhi00_drho

    sigma_kl3 = self.sigma_kl3

    Sk = self.Sk
    vki = self.vki
    vk = self.vk
    x_k = x[self.groups_index]

    xs_ki = x_k*Sk*vki*vk
    xs_m = np.sum(xs_ki)
    xs_k = xs_ki / xs_m

    xhi, dxhi_dxhi00 = xhi_eval(xhi00, xs_k, xs_m, d_kk03)
    xhix, dxhix_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, d_kl3)
    xhixm, dxhixm_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, sigma_kl3)

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3*xhix2, 4*xhix3]])

    # monomer contribution calculation
    Ckl = self.Ckl
    Ckl2 = self.Ckl2
    eps_kl = self.eps_kl
    f1, f2, f3 = self.f1, self.f2, self.f3
    f4, f5, f6 = self.f4, self.f5, self.f6
    # lambdaskl = self.lambdaskl
    ccteskl = self.ccteskl

    a1kl, a2kl = a1sB_eval(xhi00, xhix, xhix_vec[0], xs_m, I_lambdaskl,
                           J_lambdaskl, ccteskl, a1vdwkl, a1vdw_ctekl)

    # zero order pertubation
    aHS = ahs(xhi)

    # first order pertubation
    suma1_monomer = Ckl * np.sum(a1kl * x0_a1, axis=0)
    a1m = a1(xs_k, suma1_monomer)

    # second order pertubation
    khs, dkhs = dkHS_dxhi00(xhix, dxhix_dxhi00)
    suma2_monomer = Ckl2 * np.sum(a2kl * x0_a2, axis=0)
    a2m = a2(xs_k, khs, xhixm, suma2_monomer, eps_kl, f1, f2, f3)

    # third order pertubaton
    a3m = a3(xs_k, xhixm, eps_kl, f4, f5, f6)

    am = xs_m * (aHS + beta * a1m + beta**2 * a2m + beta**3 * a3m)

    # chain contribution calculation
    # lambdasii = self.lambdasii
    cctesii = self.cctesii
    alphaii = self.alphaii
    eps_ii = self.eps_ii
    Cii = self.Cii
    Cii2 = self.Cii2

    da1ii, da2ii = da1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m, I_lambdasii,
                                     J_lambdasii, cctesii, a1vdwii,
                                     a1vdw_cteii, dxhix_dxhi00)

    # g hard sphere
    gHS = gdHS(x0i_matrix, xhix)

    # gamma_c
    gc = gammac(xhixm, alphaii, tetha)

    # g1sigma
    da1_chain = Cii * np.sum(da1ii[1] * x0_a1ii, axis=0)
    suma_g1 = Cii * np.sum(da1ii[0] * x0_g1ii, axis=0)
    g1s = g1sigma(xhi00, xs_m, da1_chain, suma_g1, a1vdw_cteii)

    # g2sigma
    dsuma2_chain = Cii2 * np.sum(da2ii * x0_a2ii, axis=1)
    da2new = da2new_dxhi00(khs, dkhs, dsuma2_chain, eps_ii)
    suma_g2 = Cii2 * np.sum(da2ii[0] * x0_g2ii, axis=0)
    g2m = g2mca(xhi00, khs, xs_m, da2new, suma_g2, eps_ii, a1vdw_cteii)
    g2s = (1 + gc) * g2m

    lng = lngmie(gHS, g1s, g2s, beps_ii, beps_ii2)

    ach = - lng@(x * (self.zs_m - 1.))

    ares = am + ach

    if self.asso_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        # T_ad = 1/(self.eps_ij*beta)
        T_ad = temp_aux[29]
        sigma_kl3 = self.sigma_kl3
        sigma_x3 = np.matmul(np.matmul(sigma_kl3, xs_k), xs_k)
        rho_ad = rho * xs_m * sigma_x3

        Iijklab = np.zeros([self.nc, self.nc])
        Iab(rho_ad, T_ad, Iijklab)

        diagasso = self.diagasso
        # vki_asso = self.vki[self.group_asso_index]
        vki_asso = self.vki_asso
        DIJ = self.DIJ
        xj_asso = x[self.molecule_id_index_sites]
        xjvk = xj_asso*vki_asso

        # Fklab = np.exp(self.epsAB_kl * beta) - 1
        Fklab = temp_aux[30]
        Dijklab = self.kAB_kl * Fklab
        Dijklab[self.indexABij] *= Iijklab[self.indexAB_id]

        Xass = Xass_solver(rho, xjvk, DIJ, Dijklab, diagasso, Xass)

        ares += np.dot(self.S * xjvk, (np.log(Xass) - Xass/2 + 1/2))
    else:
        Xass = Xass0
    return ares, Xass


def dares_drho(self, x, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, d_kk, d_kl, d_kl3, d_kk03, x0_kl = temp_aux[:8]
    a1vdw_ctekl, a1vdwkl, x0_a1, x0_a2, I_lambdaskl = temp_aux[8:13]
    J_lambdaskl, d_ii, d_ii3, x0_ii, a1vdw_cteii, a1vdwii = temp_aux[13:19]
    tetha, x0_a1ii, x0_a2ii, x0_g1ii, x0_g2ii, I_lambdasii = temp_aux[19:25]
    J_lambdasii, x0i_matrix, beps_ii, beps_ii2 = temp_aux[25:29]

    dxhi00_drho = self.dxhi00_drho
    xhi00 = rho*dxhi00_drho

    sigma_kl3 = self.sigma_kl3

    Sk = self.Sk
    vki = self.vki
    vk = self.vk
    x_k = x[self.groups_index]

    xs_ki = x_k*Sk*vki*vk
    xs_m = np.sum(xs_ki)
    xs_k = xs_ki / xs_m

    xhi, dxhi_dxhi00 = xhi_eval(xhi00, xs_k, xs_m, d_kk03)
    xhix, dxhix_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, d_kl3)
    xhixm, dxhixm_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, sigma_kl3)

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3*xhix2, 4*xhix3],
                         [0., 2., 6*xhix, 12*xhix2]])

    # monomer contribution calculation
    Ckl = self.Ckl
    Ckl2 = self.Ckl2
    eps_kl = self.eps_kl
    f1, f2, f3 = self.f1, self.f2, self.f3
    f4, f5, f6 = self.f4, self.f5, self.f6
    # lambdaskl = self.lambdaskl
    ccteskl = self.ccteskl
    da1kl, da2kl = da1sB_dxhi00_eval(xhi00, xhix, xhix_vec[:2], xs_m,
                                     I_lambdaskl, J_lambdaskl, ccteskl,
                                     a1vdwkl, a1vdw_ctekl, dxhix_dxhi00)

    # zero order pertubation
    daHS = dahs_dxhi00(xhi, dxhi_dxhi00)

    # first order pertubation
    suma1_monomer = Ckl * np.sum(da1kl * x0_a1, axis=1)
    da1m = da1_dxhi00(xs_k, suma1_monomer)

    # second order pertubation
    khs, dkhs, d2khs = d2kHS_dxhi00(xhix, dxhix_dxhi00)
    suma2_monomer = Ckl2 * np.sum(da2kl * x0_a2, axis=1)
    da2m = da2_dxhi00(xs_k, khs, dkhs, xhixm, dxhixm_dxhi00, suma2_monomer,
                      eps_kl, f1, f2, f3)

    # third order pertubaton
    da3m = da3_dxhi00(xs_k, xhixm, dxhixm_dxhi00, eps_kl, f4, f5, f6)

    damono = xs_m * (daHS + beta * da1m + beta**2 * da2m + beta**3 * da3m)

    # chain contribution calculation
    # lambdasii = self.lambdasii
    cctesii = self.cctesii
    alphaii = self.alphaii
    eps_ii = self.eps_ii
    Cii = self.Cii
    Cii2 = self.Cii2

    d2a1ii, d2a2ii = d2a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m,
                                        I_lambdasii, J_lambdasii, cctesii,
                                        a1vdwii, a1vdw_cteii, dxhix_dxhi00)

    # g hard sphere
    dgHS = dgdHS_dxhi00(x0i_matrix, xhix, dxhix_dxhi00)

    # gamma_c
    dgc = dgammac_dxhi00(xhixm, dxhixm_dxhi00, alphaii, tetha)

    # g1sigma
    d2a1_chain = Cii * np.sum(d2a1ii[1:] * x0_a1ii, axis=1)
    dsuma_g1 = Cii * np.sum(d2a1ii[:2] * x0_g1ii, axis=1)
    dg1s = dg1sigma_dxhi00(xhi00, xs_m, d2a1_chain, dsuma_g1, a1vdw_cteii)

    # g2sigma
    d2suma2_chain = Cii2 * np.sum(d2a2ii * x0_a2ii, axis=1)
    d2a2new = d2a2new_dxhi00(khs, dkhs, d2khs, d2suma2_chain, eps_ii)
    dsuma_g2 = Cii2 * np.sum(d2a2ii[:2] * x0_g2ii, axis=1)
    dg2m = dg2mca_dxhi00(xhi00, khs, dkhs, xs_m, d2a2new, dsuma_g2, eps_ii,
                         a1vdw_cteii)
    dg2s = dg2m * (1 + dgc[0])
    dg2s[1] += dg2m[0] * dgc[1]

    dlng = dlngmie_dxhi00(dgHS, dg1s, dg2s, beps_ii, beps_ii2)

    dachain = - dlng@(x * (self.zs_m - 1.))
    ares = damono + dachain
    ares *= self.dxhi00_1

    if self.asso_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        # T_ad = 1/(self.eps_ij*beta)
        T_ad = temp_aux[29]
        sigma_kl3 = self.sigma_kl3
        sigma_x3 = np.matmul(np.matmul(sigma_kl3, xs_k), xs_k)
        drho_ad = xs_m * sigma_x3
        rho_ad = rho * drho_ad

        Iijklab = np.zeros([self.nc, self.nc])
        dIijklab_drho = np.zeros([self.nc, self.nc])
        dIab_drho(rho_ad, T_ad, drho_ad, Iijklab, dIijklab_drho)

        diagasso = self.diagasso
        # vki_asso = self.vki[self.group_asso_index]
        vki_asso = self.vki_asso
        DIJ = self.DIJ
        xj_asso = x[self.molecule_id_index_sites]
        xjvk = xj_asso*vki_asso

        # Fklab = np.exp(self.epsAB_kl * beta) - 1
        Fklab = temp_aux[30]
        Dijklab = self.kAB_kl * Fklab
        Dijklab[self.indexABij] *= Iijklab[self.indexAB_id]

        dDijklab_drho = self.kAB_kl * Fklab
        dDijklab_drho[self.indexABij] *= dIijklab_drho[self.indexAB_id]

        Xass = Xass_solver(rho, xjvk, DIJ, Dijklab, diagasso, Xass)
        CIJ = CIJ_matrix(rho, xjvk, Xass, DIJ, Dijklab, diagasso)
        dXass = dXass_drho(rho, xjvk, Xass, DIJ, Dijklab, dDijklab_drho, CIJ)

        ares[0] += np.dot(self.S * xjvk, (np.log(Xass) - Xass/2 + 1/2))
        ares[1] += np.dot(self.S * xjvk, (1/Xass - 1/2) * dXass)
    else:
        Xass = Xass0
    return ares, Xass


def d2ares_drho(self, x, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, d_kk, d_kl, d_kl3, d_kk03, x0_kl = temp_aux[:8]
    a1vdw_ctekl, a1vdwkl, x0_a1, x0_a2, I_lambdaskl = temp_aux[8:13]
    J_lambdaskl, d_ii, d_ii3, x0_ii, a1vdw_cteii, a1vdwii = temp_aux[13:19]
    tetha, x0_a1ii, x0_a2ii, x0_g1ii, x0_g2ii, I_lambdasii = temp_aux[19:25]
    J_lambdasii, x0i_matrix, beps_ii, beps_ii2 = temp_aux[25:29]

    dxhi00_drho = self.dxhi00_drho
    xhi00 = rho*dxhi00_drho

    sigma_kl3 = self.sigma_kl3

    Sk = self.Sk
    vki = self.vki
    vk = self.vk
    x_k = x[self.groups_index]

    xs_ki = x_k*Sk*vki*vk
    xs_m = np.sum(xs_ki)
    xs_k = xs_ki / xs_m

    xhi, dxhi_dxhi00 = xhi_eval(xhi00, xs_k, xs_m, d_kk03)
    xhix, dxhix_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, d_kl3)
    xhixm, dxhixm_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, sigma_kl3)

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3.*xhix2, 4.*xhix3],
                         [0., 2., 6*xhix, 12.*xhix2],
                         [0., 0., 6., 24.*xhix]])

    # monomer contribution calculation
    Ckl = self.Ckl
    Ckl2 = self.Ckl2
    eps_kl = self.eps_kl
    f1, f2, f3 = self.f1, self.f2, self.f3
    f4, f5, f6 = self.f4, self.f5, self.f6
    # lambdaskl = self.lambdaskl
    ccteskl = self.ccteskl

    xhi, dxhi_dxhi00 = xhi_eval(xhi00, xs_k, xs_m, d_kk03)
    xhix, dxhix_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, d_kl3)
    xhixm, dxhixm_dxhi00 = xhix_eval(xhi00, xs_k, xs_m, sigma_kl3)

    d2a1kl, d2a2kl = d2a1sB_dxhi00_eval(xhi00, xhix, xhix_vec[:3], xs_m,
                                        I_lambdaskl, J_lambdaskl, ccteskl,
                                        a1vdwkl, a1vdw_ctekl, dxhix_dxhi00)

    # zero order pertubation
    d2aHS = d2ahs_dxhi00(xhi, dxhi_dxhi00)

    # first order pertubation
    suma1_monomer = Ckl * np.sum(d2a1kl * x0_a1, axis=1)
    d2a1m = d2a1_dxhi00(xs_k, suma1_monomer)

    # second order pertubation
    khs, dkhs, d2khs, d3khs = d3kHS_dxhi00(xhix, dxhix_dxhi00)
    suma2_monomer = Ckl2 * np.sum(d2a2kl * x0_a2, axis=1)
    d2a2m = d2a2_dxhi00(xs_k, khs, dkhs, d2khs, xhixm, dxhixm_dxhi00,
                        suma2_monomer, eps_kl, f1, f2, f3)

    # third order pertubaton
    d2a3m = d2a3_dxhi00(xs_k, xhixm, dxhixm_dxhi00, eps_kl, f4, f5, f6)

    d2amono = xs_m * (d2aHS + beta * d2a1m + beta**2 * d2a2m + beta**3 * d2a3m)

    # chain contribution calculation
    # lambdasii = self.lambdasii
    cctesii = self.cctesii
    alphaii = self.alphaii
    eps_ii = self.eps_ii
    Cii = self.Cii
    Cii2 = self.Cii2

    d3a1ii, d3a2ii = d3a1sB_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m,
                                        I_lambdasii, J_lambdasii, cctesii,
                                        a1vdwii, a1vdw_cteii, dxhix_dxhi00)

    # g hard sphere
    d2gHS = d2gdHS_dxhi00(x0i_matrix, xhix, dxhix_dxhi00)

    # gamma_c
    d2gc = d2gammac_dxhi00(xhixm, dxhixm_dxhi00, alphaii, tetha)

    # g1sigma
    d3a1_chain = Cii * np.sum(d3a1ii[1:] * x0_a1ii, axis=1)
    d2suma_g1 = Cii * np.sum(d3a1ii[:3] * x0_g1ii, axis=1)
    d2g1s = d2g1sigma_dxhi00(xhi00, xs_m, d3a1_chain, d2suma_g1, a1vdw_cteii)

    # g2sigma
    d3suma2_chain = Cii2 * np.sum(d3a2ii * x0_a2ii, axis=1)
    d3a2new = d3a2new_dxhi00(khs, dkhs, d2khs, d3khs, d3suma2_chain, eps_ii)
    d2suma_g2 = Cii2 * np.sum(d3a2ii[:3] * x0_g2ii, axis=1)
    d2g2m = d2g2mca_dxhi00(xhi00, khs, dkhs, d2khs, xs_m, d3a2new, d2suma_g2,
                           eps_ii, a1vdw_cteii)
    d2g2s = d2g2m * (1. + d2gc[0])
    d2g2s[1] += d2g2m[0] * d2gc[1]
    d2g2s[2] += 2. * d2g2m[1] * d2gc[1] + d2g2m[0] * d2gc[2]

    d2lng = d2lngmie_dxhi00(d2gHS, d2g1s, d2g2s, beps_ii, beps_ii2)

    d2achain = - d2lng@(x * (self.zs_m - 1.))
    ares = d2amono + d2achain
    ares *= self.dxhi00_2

    if self.asso_bool:
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0
        # T_ad = 1/(self.eps_ij*beta)
        T_ad = temp_aux[29]
        sigma_kl3 = self.sigma_kl3
        sigma_x3 = np.matmul(np.matmul(sigma_kl3, xs_k), xs_k)
        drho_ad = xs_m * sigma_x3
        rho_ad = rho * drho_ad

        Iijklab = np.zeros([self.nc, self.nc])
        dIijklab_drho = np.zeros([self.nc, self.nc])
        d2Iijklab_drho = np.zeros([self.nc, self.nc])
        d2Iab_drho(rho_ad, T_ad, drho_ad, Iijklab, dIijklab_drho,
                   d2Iijklab_drho)

        diagasso = self.diagasso
        # vki_asso = self.vki[self.group_asso_index]
        vki_asso = self.vki_asso
        DIJ = self.DIJ
        xj_asso = x[self.molecule_id_index_sites]
        xjvk = xj_asso*vki_asso

        # Fklab = np.exp(self.epsAB_kl * beta) - 1
        Fklab = temp_aux[30]
        Dijklab = self.kAB_kl * Fklab
        Dijklab[self.indexABij] *= Iijklab[self.indexAB_id]

        dDijklab_drho = self.kAB_kl * Fklab
        dDijklab_drho[self.indexABij] *= dIijklab_drho[self.indexAB_id]

        d2Dijklab_drho = self.kAB_kl * Fklab
        d2Dijklab_drho[self.indexABij] *= d2Iijklab_drho[self.indexAB_id]

        Xass = Xass_solver(rho, xjvk, DIJ, Dijklab, diagasso, Xass)
        CIJ = CIJ_matrix(rho, xjvk, Xass, DIJ, Dijklab, diagasso)
        dXass = dXass_drho(rho, xjvk, Xass, DIJ, Dijklab, dDijklab_drho, CIJ)
        d2Xass = d2Xass_drho(rho, xjvk, Xass, dXass, DIJ, Dijklab,
                             dDijklab_drho, d2Dijklab_drho, CIJ)

        aux0 = self.S * xjvk
        aux1 = np.log(Xass) - Xass/2 + 1/2
        aux2 = 1/Xass - 1/2

        ares[0] += np.dot(aux0, aux1)
        ares[1] += np.dot(aux0, aux2 * dXass)
        ares[2] += np.dot(aux0, -(dXass/Xass)**2+d2Xass*aux2)

    else:
        Xass = Xass0
    return ares, Xass


def dares_dx(self, x, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, d_kk, d_kl, d_kl3, d_kk03, x0_kl = temp_aux[:8]
    a1vdw_ctekl, a1vdwkl, x0_a1, x0_a2, I_lambdaskl = temp_aux[8:13]
    J_lambdaskl, d_ii, d_ii3, x0_ii, a1vdw_cteii, a1vdwii = temp_aux[13:19]
    tetha, x0_a1ii, x0_a2ii, x0_g1ii, x0_g2ii, I_lambdasii = temp_aux[19:25]
    J_lambdasii, x0i_matrix, beps_ii, beps_ii2 = temp_aux[25:29]

    dxhi00_drho = self.dxhi00_drho
    xhi00 = rho*dxhi00_drho

    sigma_kl3 = self.sigma_kl3

    Sk = self.Sk
    vki = self.vki
    vk = self.vk
    dxkdx = self.dxkdx
    zs_m = self.zs_m
    x_k = x[self.groups_index]

    aux_Skvksvki = Sk*vki*vk
    xs_ki = x_k*aux_Skvksvki
    xs_m = np.sum(xs_ki)
    xs_k = xs_ki / xs_m

    dxk_dx_aux = aux_Skvksvki * dxkdx
    dxsk_dx = dxk_dx_aux * xs_m
    dxsk_dx -= np.outer(zs_m, xs_ki)
    dxsk_dx /= xs_m**2

    out = dxhi_dx_eval(xhi00, xs_k, xs_m, d_kk03, dxk_dx_aux)
    xhi, dxhi_dxhi00, dxhi_dx = out
    out = dxhix_dx_eval(xhi00, xs_k, dxsk_dx, xs_m, zs_m, d_kl3)
    xhix, dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00 = out
    out = dxhix_dx_eval(xhi00, xs_k, dxsk_dx, xs_m, zs_m, sigma_kl3)
    xhixm, dxhixm_dxhi00, dxhixm_dx, dxhixm_dx_dxhi00 = out

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3*xhix2, 4*xhix3],
                         [0., 2, 6*xhix, 12*xhix2]])

    khs, dkhs, dkhsx, dkhsxxhi = dkHS_dx_dxhi00(xhix, dxhix_dxhi00, dxhix_dx,
                                                dxhix_dx_dxhi00)

    # monomer contribution calculation
    Ckl = self.Ckl
    Ckl2 = self.Ckl2
    eps_kl = self.eps_kl
    f1, f2, f3 = self.f1, self.f2, self.f3
    f4, f5, f6 = self.f4, self.f5, self.f6
    # lambdaskl = self.lambdaskl
    ccteskl = self.ccteskl

    out = da1sB_dx_eval(xhi00, xhix, xhix_vec[:2], xs_m, zs_m, I_lambdaskl,
                        J_lambdaskl, ccteskl, a1vdwkl, a1vdw_ctekl, dxhix_dx)
    a1kl, a2kl, da1x_kl, da2x_kl = out

    # zero order pertubation
    aHS, daHSx = dahs_dx(xhi, dxhi_dx)

    # first order pertubation
    suma1_monomer = Ckl * np.sum(a1kl * x0_a1, axis=0)
    suma1x_monomer = Ckl * (da1x_kl[0]*x0_a1[0] + da1x_kl[1]*x0_a1[1])

    a1m, da1mx = da1_dx(xs_k, dxsk_dx, suma1_monomer, suma1x_monomer)

    # second order pertubation
    suma2_monomer = Ckl2 * np.sum(a2kl * x0_a2, axis=0)
    suma2x_monomer = da2x_kl[0]*x0_a2[0] + da2x_kl[1]*x0_a2[1]
    suma2x_monomer += da2x_kl[2]*x0_a2[2]
    suma2x_monomer *= Ckl2

    a2m, da2mx = da2_dx(xs_k, dxsk_dx, khs, dkhsx, xhixm, dxhixm_dx,
                        suma2_monomer, suma2x_monomer, eps_kl, f1, f2, f3)

    # third order pertubation
    a3m, da3mx = da3_dx(xs_k, dxsk_dx, xhixm, dxhixm_dx, eps_kl, f4, f5, f6)

    beta2 = beta**2
    beta3 = beta2*beta
    am = aHS + beta * a1m + beta2 * a2m + beta3 * a3m
    damx = daHSx + beta * da1mx + beta2 * da2mx + beta3 * da3mx

    amono = xs_m * am
    damonox = self.zs_m * am + xs_m * damx

    # chain contribution calculation
    # lambdasii = self.lambdasii
    cctesii = self.cctesii
    alphaii = self.alphaii
    eps_ii = self.eps_ii
    Cii = self.Cii
    Cii2 = self.Cii2

    out = da1sB_dx_dxhi00_dxxhi_eval(xhi00, xhix, xhix_vec, xs_m, zs_m,
                                     I_lambdasii, J_lambdasii, cctesii,
                                     a1vdwii, a1vdw_cteii, dxhix_dxhi00,
                                     dxhix_dx, dxhix_dx_dxhi00)
    da1ii, da2ii, da1x_ii, da2x_ii, da1_xxhi00_ii, da2_xxhi00_ii = out

    # g hard sphere
    ghs, dghsx = dgdHS_dx(x0i_matrix, xhix, dxhix_dx)

    # g1sigma
    da1_chain = Cii * np.sum(da1ii[1] * x0_a1ii, axis=0)
    da1x_chain = Cii*(da1_xxhi00_ii[0]*x0_a1ii[0]+da1_xxhi00_ii[1]*x0_a1ii[1])

    suma_g1 = Cii * np.sum(da1ii[0] * x0_g1ii, axis=0)
    suma_g1x = Cii*(da1x_ii[0] * x0_g1ii[0] + da1x_ii[1] * x0_g1ii[1])
    g1s, dg1sx = dg1sigma_dx(xhi00, xs_m, zs_m, da1_chain, da1x_chain, suma_g1,
                             suma_g1x, a1vdw_cteii)

    # gamma_c
    gc, dgcx = dgammac_dx(xhixm, dxhixm_dx, alphaii, tetha)

    # g2sigma
    suma_g2 = Cii2 * np.sum(da2ii[0] * x0_g2ii, axis=0)
    suma_g2x = da2x_ii[0]*x0_g2ii[0] + da2x_ii[1]*x0_g2ii[1]
    suma_g2x += da2x_ii[2]*x0_g2ii[2]
    suma_g2x *= Cii2

    dsuma2_chain = Cii2 * np.sum(da2ii * x0_a2ii, axis=1)

    dsuma2x_chain = da2x_ii[0] * x0_a2ii[0] + da2x_ii[1] * x0_a2ii[1]
    dsuma2x_chain += da2x_ii[2] * x0_a2ii[2]
    dsuma2x_chain *= Cii2

    dsuma2xxhi_chain = da2_xxhi00_ii[0] * x0_a2ii[0]
    dsuma2xxhi_chain += da2_xxhi00_ii[1] * x0_a2ii[1]
    dsuma2xxhi_chain += da2_xxhi00_ii[2] * x0_a2ii[2]
    dsuma2xxhi_chain *= Cii2

    da2new, da2newx = da2new_dx_dxhi00(khs, dkhs, dkhsx, dkhsxxhi,
                                       dsuma2_chain, dsuma2x_chain,
                                       dsuma2xxhi_chain, eps_ii)
    g2m, dg2mx = dg2mca_dx(xhi00, khs, dkhsx, xs_m, zs_m, da2new, da2newx,
                           suma_g2, suma_g2x, eps_ii, a1vdw_cteii)
    g2s = g2m * (1 + gc)
    dg2sx = dgcx*g2m + (1+gc)*dg2mx

    lng, dlngx = dlngmie_dx(ghs, g1s, g2s, dghsx, dg1sx, dg2sx, beps_ii,
                            beps_ii2)

    zs_m1 = (zs_m - 1.)
    xzs_m1 = x*zs_m1
    achain = - lng@xzs_m1
    dachainx = - dlngx@xzs_m1 - lng * zs_m1

    ares = amono + achain
    daresx = damonox + dachainx

    if self.asso_bool:
        nc = self.nc
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0

        # beta = temp_aux[0]
        # T_ad = 1/(self.eps_ij*beta)
        T_ad = temp_aux[29]
        aux1 = xs_k * sigma_kl3
        aux2 = np.dot(xs_k, aux1)
        sigma_x3 = np.sum(aux2)
        drho_ad = xs_m * sigma_x3
        rho_ad = rho * drho_ad
        suma1 = 2*np.sum(dxsk_dx@aux1, axis=1)
        drhoad_dx = rho * (zs_m * sigma_x3 + xs_m * suma1)

        Iijklab = np.zeros([nc, nc])
        dIijklab = np.zeros([nc, nc])
        dIab(rho_ad, T_ad, Iijklab, dIijklab)
        dIijklab_dx = np.multiply.outer(drhoad_dx, dIijklab)

        diagasso = self.diagasso
        vki_asso = self.vki[self.group_asso_index]
        DIJ = self.DIJ
        xj_asso = x[self.molecule_id_index_sites]
        xjvk = xj_asso*vki_asso
        dxjasso_dx = self.dxjasso_dx

        # Fklab = np.exp(self.epsAB_kl * beta) - 1
        Fklab = temp_aux[30]
        Dijklab = self.kAB_kl * Fklab
        Dijklab[self.indexABij] *= Iijklab[self.indexAB_id]

        dDijklab_dx = np.stack(nc*[self.kAB_kl * Fklab])
        dDijklab_dx[:, self.indexABij[0], self.indexABij[1]] *= dIijklab_dx[:, self.indexAB_id[0], self.indexAB_id[1]]

        Xass = Xass_solver(rho, xjvk, DIJ, Dijklab, diagasso, Xass)
        CIJ = CIJ_matrix(rho, xjvk, Xass, DIJ, Dijklab, diagasso)
        dXassx = dXass_dx(rho, xjvk, Xass, DIJ, Dijklab, dDijklab_dx,
                          dxjasso_dx, CIJ)

        aux1 = np.log(Xass) - Xass/2 + 1/2
        aux2 = 1/Xass - 1/2

        ares += np.dot(self.S*xjvk, aux1)
        daresx += (dxjasso_dx * aux1 + dXassx * xjvk * aux2)@self.S
    else:
        Xass = Xass0

    return ares, daresx, Xass


def dares_dx_drho(self, x, rho, temp_aux, Xass0=None):

    beta, beta2, beta3, d_kk, d_kl, d_kl3, d_kk03, x0_kl = temp_aux[:8]
    a1vdw_ctekl, a1vdwkl, x0_a1, x0_a2, I_lambdaskl = temp_aux[8:13]
    J_lambdaskl, d_ii, d_ii3, x0_ii, a1vdw_cteii, a1vdwii = temp_aux[13:19]
    tetha, x0_a1ii, x0_a2ii, x0_g1ii, x0_g2ii, I_lambdasii = temp_aux[19:25]
    J_lambdasii, x0i_matrix, beps_ii, beps_ii2 = temp_aux[25:29]

    dxhi00_drho = self.dxhi00_drho
    xhi00 = rho*dxhi00_drho

    sigma_kl3 = self.sigma_kl3

    Sk = self.Sk
    vki = self.vki
    vk = self.vk
    dxkdx = self.dxkdx
    zs_m = self.zs_m
    x_k = x[self.groups_index]

    aux_Skvksvki = Sk*vki*vk
    xs_ki = x_k*aux_Skvksvki
    xs_m = np.sum(xs_ki)
    xs_k = xs_ki / xs_m

    dxk_dx_aux = aux_Skvksvki * dxkdx
    dxsk_dx = dxk_dx_aux * xs_m
    dxsk_dx -= np.outer(zs_m, xs_ki)
    dxsk_dx /= xs_m**2

    out = dxhi_dx_eval(xhi00, xs_k, xs_m, d_kk03, dxk_dx_aux)
    xhi, dxhi_dxhi00, dxhi_dx = out
    out = dxhix_dx_eval(xhi00, xs_k, dxsk_dx, xs_m, zs_m, d_kl3)
    xhix, dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00 = out
    out = dxhix_dx_eval(xhi00, xs_k, dxsk_dx, xs_m, zs_m, sigma_kl3)
    xhixm, dxhixm_dxhi00, dxhixm_dx, dxhixm_dx_dxhi00 = out

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_vec = np.array([[xhix, xhix2, xhix3, xhix4],
                         [1., 2 * xhix, 3*xhix2, 4*xhix3],
                         [0., 2, 6*xhix, 12*xhix2]])

    out = d2kHS_dx_dxhi00(xhix, dxhix_dxhi00, dxhix_dx, dxhix_dx_dxhi00)
    khs, dkhs, d2khs, dkhsx, dkhsxxhi = out

    # monomer contribution calculation
    Ckl = self.Ckl
    Ckl2 = self.Ckl2
    eps_kl = self.eps_kl
    f1, f2, f3 = self.f1, self.f2, self.f3
    f4, f5, f6 = self.f4, self.f5, self.f6
    # lambdaskl = self.lambdaskl
    ccteskl = self.ccteskl

    out = da1sB_dx_dxhi00_eval(xhi00, xhix, xhix_vec, xs_m, zs_m,
                               I_lambdaskl, J_lambdaskl, ccteskl, a1vdwkl,
                               a1vdw_ctekl, dxhix_dxhi00, dxhix_dx)
    da1kl, da2kl, da1x_kl, da2x_kl = out

    # zero order pertubation
    aHS, daHSx = dahs_dxxhi(xhi, dxhi_dxhi00, dxhi_dx)

    # first order pertubation
    suma1_monomer = Ckl * np.sum(da1kl * x0_a1, axis=1)
    suma1x_monomer = Ckl * (da1x_kl[0]*x0_a1[0] + da1x_kl[1]*x0_a1[1])

    a1m, da1mx = da1_dxxhi(xs_k, dxsk_dx, suma1_monomer, suma1x_monomer)

    # second order pertubation
    suma2_monomer = Ckl2 * np.sum(da2kl * x0_a2, axis=1)
    suma2x_monomer = da2x_kl[0]*x0_a2[0] + da2x_kl[1]*x0_a2[1]
    suma2x_monomer += da2x_kl[2]*x0_a2[2]
    suma2x_monomer *= Ckl2

    a2m, da2mx = da2_dxxhi(xs_k, dxsk_dx, khs, dkhs, dkhsx, xhixm,
                           dxhixm_dxhi00, dxhixm_dx, suma2_monomer,
                           suma2x_monomer, eps_kl, f1, f2, f3)

    # third order pertubation
    a3m, da3mx = da3_dxxhi(xs_k, dxsk_dx, xhixm, dxhixm_dxhi00, dxhixm_dx,
                           eps_kl, f4, f5, f6)

    beta2 = beta**2
    beta3 = beta2*beta
    am = aHS + beta * a1m + beta2 * a2m + beta3 * a3m
    damx = daHSx + beta * da1mx + beta2 * da2mx + beta3 * da3mx

    amono = xs_m * am
    damonox = self.zs_m * am[0] + xs_m * damx

    # chain contribution calculation
    # lambdasii = self.lambdasii
    cctesii = self.cctesii
    alphaii = self.alphaii
    eps_ii = self.eps_ii
    Cii = self.Cii
    Cii2 = self.Cii2

    out = da1sB_dx_d2xhi00_dxxhi_eval(xhi00, xhix, xhix_vec, xs_m, zs_m,
                                      I_lambdasii, J_lambdasii, cctesii,
                                      a1vdwii, a1vdw_cteii, dxhix_dxhi00,
                                      dxhix_dx, dxhix_dx_dxhi00)
    d2a1ii, d2a2ii, da1x_ii, da2x_ii, da1_xxhi00_ii, da2_xxhi00_ii = out

    # g hard sphere
    ghs, dghsx = dgdHS_dxxhi(x0i_matrix, xhix, dxhix_dxhi00, dxhix_dx)

    # g1sigma
    d2a1_chain = Cii * np.sum(d2a1ii[1:] * x0_a1ii, axis=1)
    # da1_chain = Cii * np.sum(da1ii[1] * x0_a1ii, axis=0)
    da1x_chain = Cii*(da1_xxhi00_ii[0]*x0_a1ii[0]+da1_xxhi00_ii[1]*x0_a1ii[1])
    dsuma_g1 = Cii * np.sum(d2a1ii[:2] * x0_g1ii, axis=1)
    # suma_g1 = Cii * np.sum(da1ii[0] * x0_g1ii, axis=0)
    suma_g1x = Cii*(da1x_ii[0] * x0_g1ii[0] + da1x_ii[1] * x0_g1ii[1])
    g1s, dg1sx = dg1sigma_dxxhi(xhi00, xs_m, zs_m, d2a1_chain, da1x_chain,
                                dsuma_g1, suma_g1x, a1vdw_cteii)

    # gamma_c
    gc, dgcx = dgammac_dxxhi(xhixm, dxhixm_dxhi00, dxhixm_dx, alphaii, tetha)

    # g2sigma
    dsuma_g2 = Cii2 * np.sum(d2a2ii[:2] * x0_g2ii, axis=1)
    suma_g2x = da2x_ii[0]*x0_g2ii[0] + da2x_ii[1]*x0_g2ii[1]
    suma_g2x += da2x_ii[2]*x0_g2ii[2]
    suma_g2x *= Cii2

    dsuma2x_chain = da2x_ii[0] * x0_a2ii[0] + da2x_ii[1] * x0_a2ii[1]
    dsuma2x_chain += da2x_ii[2] * x0_a2ii[2]
    dsuma2x_chain *= Cii2

    dsuma2xxhi_chain = da2_xxhi00_ii[0] * x0_a2ii[0]
    dsuma2xxhi_chain += da2_xxhi00_ii[1] * x0_a2ii[1]
    dsuma2xxhi_chain += da2_xxhi00_ii[2] * x0_a2ii[2]
    dsuma2xxhi_chain *= Cii2

    d2suma2_chain = Cii2 * np.sum(d2a2ii * x0_a2ii, axis=1)

    *d2a2new, da2newx = da2new_dxxhi_dxhi00(khs, dkhs, d2khs, dkhsx, dkhsxxhi,
                                            d2suma2_chain, dsuma2x_chain,
                                            dsuma2xxhi_chain, eps_ii)
    g2m, dg2mx = dg2mca_dxxhi(xhi00, khs, dkhs, dkhsx, xs_m, zs_m, d2a2new,
                              da2newx, dsuma_g2, suma_g2x, eps_ii, a1vdw_cteii)
    g2s = g2m * (1 + gc[0])
    g2s[1] += g2m[0] * gc[1]
    dg2sx = dgcx*g2m[0] + (1 + gc[0])*dg2mx

    lng, dlngx = dlngmie_dxxhi(ghs, g1s, g2s, dghsx, dg1sx, dg2sx, beps_ii,
                               beps_ii2)

    zs_m1 = (zs_m - 1.)
    xzs_m1 = x*zs_m1
    achain = - lng@xzs_m1
    dachainx = - dlngx@xzs_m1 - lng[0] * zs_m1

    ares = amono + achain
    ares *= self.dxhi00_1
    daresx = damonox + dachainx

    if self.asso_bool:
        nc = self.nc
        if Xass0 is None:
            Xass = 0.2 * np.ones(self.nsites)
        else:
            Xass = 1. * Xass0

        # beta = temp_aux[0]
        # T_ad = 1/(self.eps_ij*beta)
        T_ad = temp_aux[29]
        aux1 = xs_k * sigma_kl3
        aux2 = np.dot(xs_k, aux1)
        sigma_x3 = np.sum(aux2)
        drho_ad = xs_m * sigma_x3
        rho_ad = rho * drho_ad
        suma1 = 2*np.sum(dxsk_dx@aux1, axis=1)
        drhoad_dx = rho * (zs_m * sigma_x3 + xs_m * suma1)

        Iijklab = np.zeros([nc, nc])
        dIijklab = np.zeros([nc, nc])
        dIab(rho_ad, T_ad, Iijklab, dIijklab)
        dIijklab_dx = np.multiply.outer(drhoad_dx, dIijklab)
        dIijklab_drho = dIijklab*drho_ad

        diagasso = self.diagasso
        vki_asso = self.vki[self.group_asso_index]
        DIJ = self.DIJ
        xj_asso = x[self.molecule_id_index_sites]
        xjvk = xj_asso*vki_asso
        dxjasso_dx = self.dxjasso_dx

        # Fklab = np.exp(self.epsAB_kl * beta) - 1
        Fklab = temp_aux[30]
        Dijklab = self.kAB_kl * Fklab
        Dijklab[self.indexABij] *= Iijklab[self.indexAB_id]

        dDijklab_drho = self.kAB_kl * Fklab
        dDijklab_drho[self.indexABij] *= dIijklab_drho[self.indexAB_id]

        dDijklab_dx = np.stack(nc*[self.kAB_kl * Fklab])
        dDijklab_dx[:, self.indexABij[0], self.indexABij[1]] *= dIijklab_dx[:, self.indexAB_id[0], self.indexAB_id[1]]

        Xass = Xass_solver(rho, xjvk, DIJ, Dijklab, diagasso, Xass)
        CIJ = CIJ_matrix(rho, xjvk, Xass, DIJ, Dijklab, diagasso)
        dXass = dXass_drho(rho, xjvk, Xass, DIJ, Dijklab, dDijklab_drho, CIJ)
        dXassx = dXass_dx(rho, xjvk, Xass, DIJ, Dijklab, dDijklab_dx,
                          dxjasso_dx, CIJ)

        aux1 = np.log(Xass) - Xass/2 + 1/2
        aux2 = 1/Xass - 1/2

        ares[0] += np.dot(self.S*xjvk, aux1)
        ares[1] += np.dot(self.S*xjvk, aux2 * dXass)
        daresx += (dxjasso_dx * aux1 + dXassx * xjvk * aux2)@self.S
    else:
        Xass = Xass0

    return ares, daresx, Xass



from __future__ import division, print_function, absolute_import
import numpy as np


def a1(xs, a1kl):
    a = np.matmul(np.matmul(a1kl, xs), xs)
    return a


def da1_dxhi00(xs, da1kl):
    da = np.matmul(np.matmul(da1kl, xs), xs)
    return da


def d2a1_dxhi00(xs, d2a1kl):
    d2a = np.matmul(np.matmul(d2a1kl, xs), xs)
    return d2a


def d3a1_dxhi00(xs, d3a1kl):
    d3a = np.matmul(np.matmul(d3a1kl, xs), xs)
    return d3a



from __future__ import division, print_function, absolute_import
import numpy as np
from ..math import gauss


def fit_cii(tension, Tsat, Psat, rhol, rhov, eos, n=100, deg=0):
    """
    fit_cii
    Optimize influence parameter for SGT for pure componentes


    Parameters
    ----------
    tension : array
        Experimental interfacial tension array
    Tsat : array
        Experimental saturation temperature [K]
    Psat : array
        Computed saturation pressure [Pa]
    rhol : array
        Computed liquid density [mol/m3]
    rhov : array
        Computed vapor density [mol/m3]
    eos : object
        created from pure fluid and saftvrmie function
    deg : int
        degree of the influence parameter polynomial

    Returns
    -------
    cii : float
        Influence Parameter of the fluid
    """
    n = 100
    roots, weigths = gauss(n)
    tena = np.zeros_like(Tsat)
    ndata = len(tension)
    for i in range(ndata):
        Tfactor, Pfactor, rofactor, tenfactor = eos.sgt_adim_fit(Tsat[i])
        rola = rhol[i] * rofactor
        rova = rhov[i] * rofactor
        Tad = Tsat[i] * Tfactor
        Pad = Psat[i] * Pfactor
        # Equilibrium chemical potential
        mu0 = eos.muad(rova, Tad)

        roi = (rola-rova) * roots + rova
        wreal = (rola-rova) * weigths
        dOm = np.zeros(n)
        for j in range(n):
            dOm[j] = eos.dOm(roi[j], Tad, mu0, Pad)
        tenint = np.nan_to_num(np.sqrt(2*dOm))
        tena[i] = np.dot(wreal, tenint)
        tena[i] *= tenfactor

    cii0 = (tension/tena)**2
    cii = np.polyfit(Tsat, cii0, deg)
    return cii



from __future__ import division, print_function, absolute_import
import numpy as np
from .fitmulticomponent import fobj_vle, fobj_lle, fobj_vlleb
from scipy.optimize import minimize, minimize_scalar
from ..saft import saftvrmie


def fobj_kij(kij, mix, datavle=None, datalle=None, datavlle=None,
             weights_vle=[1., 1.], weights_lle=[1., 1.],
             weights_vlle=[1., 1., 1., 1.]):

    Kij = np.array([[0, kij], [kij, 0]])
    mix.kij_saft(Kij)
    eos = saftvrmie(mix)

    error = 0.
    if datavle is not None:
        error += fobj_vle(eos, *datavle, weights_vle=weights_vle)
    if datalle is not None:
        error += fobj_lle(eos, *datalle, weights_lle=weights_lle)
    if datavlle is not None:
        error += fobj_vlleb(eos, *datavlle, weights_vlleb=weights_vlle)
    return error


def fit_kij(kij_bounds, mix, datavle=None, datalle=None, datavlle=None,
            weights_vle=[1., 1.], weights_lle=[1., 1.],
            weights_vlle=[1., 1., 1., 1.], minimize_options={}):
    """
    fit_kij attemps to fit kij to VLE, LLE, VLLE

    Parameters
    ----------
    kij_bounds: tuple
        bounds for kij correction
    mix: object
        binary mixture
    datavle: tuple, optional
        (Xexp, Yexp, Texp, Pexp)
    datalle: tuple, optional
        (Xexp, Wexp, Texp, Pexp)
    datavlle: tuple, optional
        (Xexp, Wexp, Yexp, Texp, Pexp)
    weights_vle: list or array_like, optional
        weights_vle[0] = weight for Y composition error, default to 1.
        weights_vle[1] = weight for bubble pressure error, default to 1.
    weights_lle: list or array_like, optional
        weights_lle[0] = weight for X (liquid 1) composition error, default to 1.
        weights_lle[1] = weight for W (liquid 2) composition error, default to 1.
    weights_vlle: list or array_like, optional
        weights_vlle[0] = weight for X (liquid 1) composition error, default to 1.
        weights_vlle[1] = weight for W (liquid 2) composition error, default to 1.
        weights_vlle[2] = weight for Y (vapor) composition error, default to 1.
        weights_vlle[3] = weight for equilibrium pressure error, default to 1.
     minimize_options: dict
        Dictionary of any additional spefication for scipy minimize_scalar

    Returns
    -------
    fit : OptimizeResult
        Result of SciPy minimize

    """
    fit = minimize_scalar(fobj_kij, kij_bounds, args=(mix, datavle, datalle,
                          datavlle, weights_vle, weights_lle, weights_vlle),
                          **minimize_options)
    return fit


def fobj_asso(x, mix, datavle=None, datalle=None, datavlle=None,
              weights_vle=[1., 1.], weights_lle=[1., 1.],
              weights_vlle=[1., 1., 1., 1.]):
    kij, lij = x
    Kij = np.array([[0, kij], [kij, 0]])
    mix.kij_saft(Kij)
    Lij = np.array([[0, lij], [lij, 0]])
    mix.lij_saft(Lij)
    eos = saftvrmie(mix)

    error = 0.
    if datavle is not None:
        error += fobj_vle(eos, *datavle, weights_vle=weights_vle)
    if datalle is not None:
        error += fobj_lle(eos, *datalle, weights_lle=weights_lle)
    if datavlle is not None:
        error += fobj_vlleb(eos, *datavlle, weights_vlleb=weights_vlle)
    return error


def fit_asso(x0, mix, datavle=None, datalle=None, datavlle=None,
             weights_vle=[1., 1.], weights_lle=[1., 1.],
             weights_vlle=[1., 1., 1., 1.], minimize_options={}):
    """
    fit_asso attemps to fit kij and lij to VLE, LLE, VLLE

    Parameters
    ----------
    x0: array
        initial values for kij and lij
    mix: object
        binary mixture
    datavle: tuple, optional
        (Xexp, Yexp, Texp, Pexp)
    datalle: tuple, optional
        (Xexp, Wexp, Texp, Pexp)
    datavlle: tuple, optional
        (Xexp, Wexp, Yexp, Texp, Pexp)
    weights_vle: list or array_like, optional
        weights_vle[0] = weight for Y composition error, default to 1.
        weights_vle[1] = weight for bubble pressure error, default to 1.
    weights_lle: list or array_like, optional
        weights_lle[0] = weight for X (liquid 1) composition error, default to 1.
        weights_lle[1] = weight for W (liquid 2) composition error, default to 1.
    weights_vlle: list or array_like, optional
        weights_vlle[0] = weight for X (liquid 1) composition error, default to 1.
        weights_vlle[1] = weight for W (liquid 2) composition error, default to 1.
        weights_vlle[2] = weight for Y (vapor) composition error, default to 1.
        weights_vlle[3] = weight for equilibrium pressure error, default to 1.
     minimize_options: dict
        Dictionary of any additional spefication for scipy minimize

    Returns
    -------
    fit : OptimizeResult
        Result of SciPy minimize

    """
    fit = minimize(fobj_asso, x0, args=(mix, datavle, datalle, datavlle,
                   weights_vle, weights_lle, weights_vlle),
                   **minimize_options)
    return fit


def fobj_cross(x, mix, assoc, datavle=None, datalle=None, datavlle=None,
               weights_vle=[1., 1.], weights_lle=[1., 1.],
               weights_vlle=[1., 1., 1., 1.]):
    kij, rc = x
    Kij = np.array([[0, kij], [kij, 0]])
    mix.kij_saft(Kij)

    eos = saftvrmie(mix)
    eos.eABij[0, 1] = mix.eAB[assoc] / 2
    eos.eABij[1, 0] = mix.eAB[assoc] / 2
    eos.rcij[0, 1] = rc * 1e-10
    eos.rcij[1, 0] = rc * 1e-10

    error = 0.
    if datavle is not None:
        error += fobj_vle(eos, *datavle, weights_vle=weights_vle)
    if datalle is not None:
        error += fobj_lle(eos, *datalle, weights_lle=weights_lle)
    if datavlle is not None:
        error += fobj_vlleb(eos, *datavlle, weights_vlleb=weights_vlle)
    return error


def fit_cross(x0, mix, assoc, datavle=None, datalle=None, datavlle=None,
              weights_vle=[1., 1.], weights_lle=[1., 1.],
              weights_vlle=[1., 1., 1., 1.], minimize_options={}):
    """
    fit_cross attemps to fit kij and rcij to VLE, LLE, VLLE

    Parameters
    ----------
    x0 : array
        initial values for kij and rc
    mix: object
        binary mixture
    assoc : int
        index of associating component
    datavle: tuple, optional
        (Xexp, Yexp, Texp, Pexp)
    datalle: tuple, optional
        (Xexp, Wexp, Texp, Pexp)
    datavlle: tuple, optional
        (Xexp, Wexp, Yexp, Texp, Pexp)
    weights_vle: list or array_like, optional
        weights_vle[0] = weight for Y composition error, default to 1.
        weights_vle[1] = weight for bubble pressure error, default to 1.
    weights_lle: list or array_like, optional
        weights_lle[0] = weight for X (liquid 1) composition error, default to 1.
        weights_lle[1] = weight for W (liquid 2) composition error, default to 1.
    weights_vlle: list or array_like, optional
        weights_vlle[0] = weight for X (liquid 1) composition error, default to 1.
        weights_vlle[1] = weight for W (liquid 2) composition error, default to 1.
        weights_vlle[2] = weight for Y (vapor) composition error, default to 1.
        weights_vlle[3] = weight for equilibrium pressure error, default to 1.
    minimize_options: dict
        Dictionary of any additional spefication for scipy minimize

    Returns
    -------
    fit : OptimizeResult
        Result of SciPy minimize

    """
    fit = minimize(fobj_cross, x0, args=(mix, assoc, datavle, datalle,
                   datavlle, weights_vle, weights_lle, weights_vlle),
                   **minimize_options)
    return fit



from __future__ import division, print_function, absolute_import
import numpy as np


# Third pertubation
def a3(xs, xhixm, eps_kl, f4, f5, f6):
    a3kl = np.exp(f5 * xhixm + f6 * xhixm**2)
    a3kl *= -eps_kl**3 * f4 * xhixm
    a = np.dot(xs, np.dot(a3kl, xs))
    return a


def da3_dxhi00(xs, xhixm, dxhixm_dxhi00, eps_kl, f4, f5, f6):

    cte = -eps_kl**3 * f4 * np.exp(f5 * xhixm + f6 * xhixm**2)

    a3kl = cte * xhixm

    da3kl = cte * (1 + f5 * xhixm + 2 * f6 * xhixm**2) * dxhixm_dxhi00

    a3 = np.array([a3kl, da3kl])
    a = np.matmul(np.matmul(a3, xs), xs)

    return a


def d2a3_dxhi00(xs, xhixm, dxhim_dxhi00, eps_kl, f4, f5, f6):

    cte = -eps_kl**3 * f4 * np.exp(f5 * xhixm + f6 * xhixm**2)

    a3kl = cte * xhixm

    da3kl = cte * (1 + f5 * xhixm + 2 * f6 * xhixm**2) * dxhim_dxhi00

    d2a3kl = cte * dxhim_dxhi00**2

    aux3 = 2 * f5 + f5**2 * xhixm + 6 * f6 * xhixm
    aux3 += 4 * f5 * f6 * xhixm**2 + 4 * f6**2*xhixm**3
    d2a3kl *= aux3
    a3 = np.array([a3kl, da3kl, d2a3kl])
    a = np.matmul(np.matmul(a3, xs), xs)

    return a



from __future__ import division, print_function, absolute_import
import numpy as np
from ..constants import kb, Na
from ..math import gauss
from .a1sB_monomer import x0lambda_evalm, x0lambda_evalc
from .monomer_aux import I_lam, J_lam
from .ares import ares, dares_drho, d2ares_drho
from .ideal import aideal, daideal_drho, d2aideal_drho
from .density_solver import density_topliss, density_newton
from .psat_saft import psat
from .tsat_saft import tsat
from .critical_pure import get_critical
from .association_aux import association_solver, association_check

R = kb * Na


def U_mie(r, c, eps, lambda_r, lambda_a):
    u = c * eps * (np.power.outer(r, lambda_r) - np.power.outer(r, lambda_a))
    return u


# Second perturbation
phi16 = np.array([[7.5365557, -37.60463, 71.745953, -46.83552, -2.467982,
                 -0.50272, 8.0956883],
                 [-359.44, 1825.6, -3168.0, 1884.2, -0.82376, -3.1935, 3.7090],
                 [1550.9, -5070.1, 6534.6, -3288.7, -2.7171, 2.0883, 0],
                 [-1.19932, 9.063632, -17.9482, 11.34027, 20.52142, -56.6377,
                 40.53683],
                 [-1911.28, 21390.175, -51320.7, 37064.54, 1103.742, -3264.61,
                 2556.181],
                 [9236.9, -129430., 357230., -315530., 1390.2, -4518.2,
                 4241.6]])


nfi = np.arange(0, 7)
nfi_num = nfi[:4]
nfi_den = nfi[4:]


# Equation A26
def fi(alphaij, i):
    phi = phi16[i-1]
    num = np.tensordot(np.power.outer(alphaij, nfi_num), phi[nfi_num],
                       axes=(-1, -1))
    den = 1 + np.tensordot(np.power.outer(alphaij, nfi_den - 3), phi[nfi_den],
                           axes=(-1, -1))
    return num/den


class saftgammamie_pure():
    '''
    Pure component SAFT-Gamma-Mie EoS Object

    This object have implemeted methods for phase equilibrium
    as for interfacial properties calculations.

    Parameters
    ----------
    pure : object
        pure component created with component class
    compute_critical: bool
        If True the critical point of the fluid will attempt to be computed
        (it might fail for some fluids).

    Attributes
    ----------
    vki: array, number of groups i
    subgroups: array, groups present in the fluid
    ngroups: int, number of groups in the fluid

    vk: array, number of sites of each group
    Sk: array, shape factor o of each group

    lr_kl: array, repulsive exponent used in monomer term
    la_kl: array, attractive exponent used in monomer term
    lar_kl: array, sum of repulsive and attractive exponents
    eps_kl: array, energy interactions used in monomer term [J]
    sigma_kl: array, lenght scale used in monomer term [m]
    Ckl: array, prefactor of mie potential used in monomer term
    alphakl: array, van der waals constant used in monomer term

    lr_kk: array, repulsive exponent used in monomer term
    la_kk: array, attractive exponent used in monomer term
    lar_kk: array, sum of repulsive and attractive exponents
    eps_kk: array, energy interactions used in monomer term [J]
    sigma_kk: array, lenght scale used in monomer term [m]
    Ckk: array, prefactor of mie potential used in monomer term
    alphakk: array, van der waals constant used in monomer term

    xs_k: array, fraction of group k
    xs_m: float, equivalent chain lenght

    zs_k: array, fraction of group k in each molecule

    lr_ii: array, repulsive exponent used in chain term
    la_ii: array, attractive exponent used in chain term
    lar_ii: array, sum of repulsive and attractive exponents
    eps_ii: array, energy interactions used in chain term [J]
    sigma_ii: array, lenght scale used in chain term [m]
    Cii: array, prefactor of mie potential used in chain term
    alphaii: array, van der waals constant used in chain term

    eAB_kl: array, association energy [J]
    kAB_kl: array, association volume [m]


    cii : influence factor for SGT [J m^5 / mol^2]

    Methods
    -------
    diameter : computes the diameter at given temperature
    temperature_aux : computes temperature dependent parameters of the fluid
    density : computes the density of the fluid
    psat : computes saturation pressure
    tsat : computes saturation temperature
    get_critical : attemps to compute the critical point of the fluid

    ares: computes the residual energy of the fluid
    dares_drho : computes residual Helmholtz energy and its density derivative
    d2ares_drho : computes residual Helmholtz energy and it density derivatives
    afcn: computes total Helmholtz energy
    dafcn_drho : computes total Helmholtz energy and its density derivative
    d2afcn_drho : computes total Helmholtz energy and it density derivatives
    pressure : computes the pressure
    dP_drho : computes pressure and its density derivative
    logfug : computes the fugacity coefficient
    a0ad : computes dimensionless Helmholtz density energy
    muad : computes dimensionless chemical potential
    dOm : computes dimensionless Thermodynamic Grand Potential
    ci :  computes influence parameters matrix for SGT
    sgt_adim : computes dimensionless factors for SGT

    association_solver: computes the fraction of non-bonded sites
    association_check: checks the fraction of non-bonded sites solution

    EntropyR : computes the residual entropy of the fluid
    EnthalpyR : computes the residual enthalpy of the fluid
    CvR : computes the residual isochoric heat capacity
    CpR : computes the residual heat capacity
    speed_sound : computes the speed of sound

    Auxiliar methods (computed using temperature_aux output list)
    -------------------------------------------------------------
    density_aux : computes density
    afcn_aux : computes afcn
    dafcn_aux : computes dafcn_drho
    d2afcn_aux : computes d2afcn_drho
    pressure_aux : computes pressure
    dP_drho_aux : computes dP_drho
    logfug_aux : computes logfug
    a0ad_aux : compute a0ad
    muad_aux : computes muad
    dOm_aux : computes dOm
    '''
    def __init__(self, pure, compute_critical=True):

        self.mw_kk = pure.mw_kk
        self.Mw = pure.Mw

        self.vki = pure.vki
        self.subgroups = pure.subgroups
        self.ngroups = pure.ngroups
        self.groups_index = pure.groups_index
        self.groups_indexes = pure.groups_indexes

        self.vk = pure.vk
        self.Sk = pure.Sk

        self.lr_kl = pure.lr_kl
        self.la_kl = pure.la_kl
        self.lar_kl = self.lr_kl + self.la_kl
        self.eps_kl = pure.eps_kl * kb
        self.sigma_kl = pure.sigma_kl * 1e-10
        self.sigma_kl3 = self.sigma_kl**3

        self.lr_kk = pure.lr_kk
        self.la_kk = pure.la_kk
        self.lar_kk = self.lr_kk + self.la_kk
        self.eps_kk = pure.eps_kk * kb
        self.sigma_kk = pure.sigma_kk * 1e-10
        self.sigma_kk3 = self.sigma_kk**3

        lr_kl, la_kl = self.lr_kl, self.la_kl
        lar_kl = lr_kl + la_kl
        dif_ckl = lr_kl - la_kl
        self.Ckl = lr_kl/dif_ckl*(lr_kl/la_kl)**(la_kl/dif_ckl)
        self.alphakl = self.Ckl * (1./(la_kl - 3.) - 1./(lr_kl - 3.))

        self.diag_index = np.diag_indices(self.ngroups)

        self.Ckk = self.Ckl[self.diag_index]
        self.alphakk = self.alphakl[self.diag_index]

        c_matrix = np.array([[0.81096, 1.7888, -37.578, 92.284],
                            [1.0205, -19.341, 151.26, -463.5],
                            [-1.9057, 22.845, -228.14, 973.92],
                            [1.0885, -6.1962, 106.98, -677.64]])

        lrkl_power = np.power.outer(lr_kl, [0, -1, -2, -3])
        cctes_lrkl = np.tensordot(lrkl_power, c_matrix, axes=(-1, -1))

        lakl_power = np.power.outer(la_kl, [0, -1, -2, -3])
        cctes_lakl = np.tensordot(lakl_power, c_matrix, axes=(-1, -1))

        lrkl_2power = np.power.outer(2*lr_kl, [0, -1, -2, -3])
        cctes_2lrkl = np.tensordot(lrkl_2power, c_matrix, axes=(-1, -1))

        lakl_2power = np.power.outer(2*la_kl, [0, -1, -2, -3])
        cctes_2lakl = np.tensordot(lakl_2power, c_matrix, axes=(-1, -1))

        larkl_power = np.power.outer(lar_kl, [0, -1, -2, -3])
        cctes_larkl = np.tensordot(larkl_power, c_matrix, axes=(-1, -1))

        # Monomer necessary term
        self.lambdaskl = (la_kl, lr_kl, 2*la_kl, 2*lr_kl, lar_kl)
        self.ccteskl = (cctes_lakl, cctes_lrkl, cctes_2lakl, cctes_2lrkl,
                        cctes_larkl)

        # For second and third order perturbation
        self.f1 = fi(self.alphakl, 1)
        self.f2 = fi(self.alphakl, 2)
        self.f3 = fi(self.alphakl, 3)
        self.f4 = fi(self.alphakl, 4)
        self.f5 = fi(self.alphakl, 5)
        self.f6 = fi(self.alphakl, 6)

        Sk = self.Sk
        vki = self.vki
        vk = self.vk

        self.xs_ki = Sk*vki*vk
        self.xs_m = np.sum(self.xs_ki)
        self.xs_k = self.xs_ki / self.xs_m

        # Chain needed parameters
        zs_ki = Sk*vki*vk
        zs_m = np.sum(zs_ki)
        zs_k = zs_ki / zs_m

        self.zs_k = zs_k

        self.eps_ii = np.matmul(np.matmul(zs_k, self.eps_kl), zs_k)
        self.la_ii = np.matmul(np.matmul(zs_k, la_kl), zs_k)
        self.lr_ii = np.matmul(np.matmul(zs_k, lr_kl), zs_k)
        self.lar_ii = self.la_ii + self.lr_ii
        self.sigma_ii3 = np.matmul(np.matmul(zs_k, self.sigma_kl3), zs_k)
        self.sigma_ii = np.cbrt(self.sigma_ii3)

        lr_ii, la_ii, lar_ii = self.lr_ii, self.la_ii, self.lar_ii
        dif_cii = lr_ii - la_ii
        self.Cii = lr_ii/dif_cii*(lr_ii/la_ii)**(la_ii/dif_cii)
        self.alphaii = self.Cii * (1./(la_ii - 3.) - 1./(lr_ii - 3.))
        self.Cii2 = self.Cii*2

        lrii_power = np.power.outer(lr_ii, [0, -1, -2, -3])
        cctes_lrii = np.tensordot(lrii_power, c_matrix, axes=(-1, -1))

        laii_power = np.power.outer(la_ii, [0, -1, -2, -3])
        cctes_laii = np.tensordot(laii_power, c_matrix, axes=(-1, -1))

        lrii_2power = np.power.outer(2*lr_ii, [0, -1, -2, -3])
        cctes_2lrii = np.tensordot(lrii_2power, c_matrix, axes=(-1, -1))

        laii_2power = np.power.outer(2*la_ii, [0, -1, -2, -3])
        cctes_2laii = np.tensordot(laii_2power, c_matrix, axes=(-1, -1))

        larii_power = np.power.outer(lar_ii, [0, -1, -2, -3])
        cctes_larii = np.tensordot(larii_power, c_matrix, axes=(-1, -1))

        # Chain necessary term
        self.lambdasii = (la_ii, lr_ii, 2*la_ii, 2*lr_ii, lar_ii)
        self.cctesii = (cctes_laii, cctes_lrii, cctes_2laii, cctes_2lrii,
                        cctes_larii)

        # For diameter calculation
        roots, weights = gauss(30)
        self.roots = roots
        self.weights = weights
        self.umie = U_mie(1./roots, self.Ckk, self.eps_kk, self.lr_kk,
                          self.la_kk)

        self.dxhi00_drho = np.pi / 6
        self.dxhi00_1 = np.array([1., self.dxhi00_drho])
        self.dxhi00_2 = np.array([1., self.dxhi00_drho, self.dxhi00_drho**2])

        self.asso_bool = pure.asso_bool
        if self.asso_bool:
            self.S = pure.S
            self.sites_asso = pure.sites_asso
            self.kAB_kl = pure.kAB_kl * 1e-30
            self.epsAB_kl = pure.epsAB_kl * kb
            self.DIJ = pure.DIJ
            self.diagasso = pure.diagasso
            self.nsites = pure.nsites
            # self.molecule_id_index_asso = pure.molecule_id_index_asso
            self.group_asso_index = pure.group_asso_index
            # self.group_id_asso = pure.group_id_asso
            # self.n_sites_molecule = pure.n_sites_molecule

        # for SGT calculation
        self.cii = pure.cii

        # computing critical point
        self.critical = False
        if compute_critical:
            out = get_critical(self, None, None, method='hybr',
                               full_output=True)
            if out.success:
                self.critical = True
                self.Tc = out.Tc
                self.Pc = out.Pc
                self.rhoc = out.rhoc

    def diameter(self, beta):
        integrer = np.exp(-beta * self.umie)
        d = self.sigma_kk * (1. - np.matmul(self.weights, integrer))
        return d

    def temperature_aux(self, T):
        """
        temperature_aux(T)

        Method that computes temperature dependent parameters.
        It returns the following list:

        Parameters
        ----------

        T : float
            Absolute temperature [K]

        Returns
        -------
        temp_aux : list
             list of computed parameters
        """

        beta = 1 / (kb * T)
        beta2 = beta**2
        beta3 = beta2*beta
        d_kk = self.diameter(beta)
        d_kl = np.add.outer(d_kk, d_kk)/2
        d_kl3 = d_kl**3
        d_kk03 = np.power.outer(d_kk, np.arange(4))

        x0_kl = self.sigma_kl / d_kl

        # for monomer contribution calculation
        a1vdw_ctekl = -12. * self.eps_kl * d_kl3

        la_kl, lr_kl, lar_kl = self.la_kl, self.lr_kl, self.lar_kl
        a1vdw_lakl = a1vdw_ctekl / (la_kl - 3)
        a1vdw_lrkl = a1vdw_ctekl / (lr_kl - 3)
        a1vdw_2lakl = a1vdw_ctekl / (2*la_kl - 3)
        a1vdw_2lrkl = a1vdw_ctekl / (2*lr_kl - 3)
        a1vdw_larkl = a1vdw_ctekl / (lar_kl - 3)
        a1vdwkl = (a1vdw_lakl, a1vdw_lrkl, a1vdw_2lakl, a1vdw_2lrkl,
                   a1vdw_larkl)

        x0_a1, x0_a2 = x0lambda_evalm(x0_kl, la_kl, lr_kl, lar_kl)
        # I and J used B term in monomer contritubion
        I_lakl = I_lam(x0_kl, la_kl)
        I_lrkl = I_lam(x0_kl, lr_kl)
        I_2lakl = I_lam(x0_kl, 2*la_kl)
        I_2lrkl = I_lam(x0_kl, 2*lr_kl)
        I_larkl = I_lam(x0_kl, lar_kl)
        I_lambdaskl = (I_lakl, I_lrkl, I_2lakl, I_2lrkl, I_larkl)

        J_lakl = J_lam(x0_kl, la_kl)
        J_lrkl = J_lam(x0_kl, lr_kl)
        J_2lakl = J_lam(x0_kl, 2*la_kl)
        J_2lrkl = J_lam(x0_kl, 2*lr_kl)
        J_larkl = J_lam(x0_kl, lar_kl)
        J_lambdaskl = (J_lakl, J_lrkl, J_2lakl, J_2lrkl, J_larkl)

        # for chain contribution calculation
        d_ii3 = np.matmul(np.matmul(self.zs_k, d_kl3), self.zs_k)
        d_ii = np.cbrt(d_ii3)
        x0_ii = self.sigma_ii/d_ii

        a1vdw_cteii = -12. * self.eps_ii * d_ii3
        la_ii, lr_ii, lar_ii = self.la_ii, self.lr_ii, self.lar_ii
        a1vdw_laii = a1vdw_cteii / (la_ii - 3)
        a1vdw_lrii = a1vdw_cteii / (lr_ii - 3)
        a1vdw_2laii = a1vdw_cteii / (2*la_ii - 3)
        a1vdw_2lrii = a1vdw_cteii / (2*lr_ii - 3)
        a1vdw_larii = a1vdw_cteii / (lar_ii - 3)
        a1vdwii = (a1vdw_laii, a1vdw_lrii, a1vdw_2laii, a1vdw_2lrii,
                   a1vdw_larii)
        # for a1 and a2 pertubation evaluation in chain contributon
        out = x0lambda_evalc(x0_ii, la_ii, lr_ii, lar_ii)
        x0_a1ii, x0_a2ii, x0_g1ii, x0_g2ii = out

        # I and J used B term in monomer contritubion
        I_laii = I_lam(x0_ii, la_ii)
        I_lrii = I_lam(x0_ii, lr_ii)
        I_2laii = I_lam(x0_ii, 2*la_ii)
        I_2lrii = I_lam(x0_ii, 2*lr_ii)
        I_larii = I_lam(x0_ii, lar_ii)
        I_lambdasii = (I_laii, I_lrii, I_2laii, I_2lrii, I_larii)

        J_laii = J_lam(x0_ii, la_ii)
        J_lrii = J_lam(x0_ii, lr_ii)
        J_2laii = J_lam(x0_ii, 2*la_ii)
        J_2lrii = J_lam(x0_ii, 2*lr_ii)
        J_larii = J_lam(x0_ii, lar_ii)
        J_lambdasii = (J_laii, J_lrii, J_2laii, J_2lrii, J_larii)

        # used in gdHS in chain contribution
        x0i_matrix = np.array([x0_ii**0, x0_ii, x0_ii**2, x0_ii**3])

        # for gamma_c in chain contribution
        beps_ii = self.eps_ii*beta
        beps_ii2 = beps_ii**2

        # tetha = np.exp(beta * self.eps_ii) - 1.
        tetha = np.exp(beps_ii) - 1.

        temp_aux = [beta, beta2, beta3, d_kk, d_kl, d_kl3, d_kk03, x0_kl,
                    a1vdw_ctekl, a1vdwkl, x0_a1, x0_a2, I_lambdaskl,
                    J_lambdaskl, d_ii, d_ii3, x0_ii, a1vdw_cteii, a1vdwii,
                    tetha, x0_a1ii, x0_a2ii, x0_g1ii, x0_g2ii, I_lambdasii,
                    J_lambdasii, x0i_matrix, beps_ii, beps_ii2]

        if self.asso_bool:
            # T_ad = 1/(self.eps_ii*beta)
            T_ad = 1/beps_ii
            Fklab = np.exp(self.epsAB_kl * beta) - 1
            temp_aux.append(T_ad)
            temp_aux.append(Fklab)
        return temp_aux

    def association_solver(self, rho, T, Xass0=None):
        """
        association_solver(rho, T, Xass0)
        Method that computes the fraction of non-bonded sites.

        Parameters
        ----------
        rho: float
            molecular density [mol/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        Xass: array
            fraction of non-bonded sites. (None if the fluid is non-associating)
        """

        if self.asso_bool:
            temp_aux = self.temperature_aux(T)
            rhom = rho * Na
            Xass = association_solver(self, rhom, temp_aux, Xass0)
        else:
            Xass = None

        return Xass

    def association_check(self, rho, T, Xass):
        """
        association_check(rho, T, Xass)
        Method that checks the fraction of non-bonded sites.

        Parameters
        ----------
        rho: float
            molecular density [mol/m3]
        T: float
            absolute temperature [K]
        Xass: array
            fraction of non-bonded sites

        Returns
        -------
        fo: array
            objective function that should be zero if the association sites
            are consistent. (Zero if the fluid is non-associating)
        """

        if self.asso_bool:
            temp_aux = self.temperature_aux(T)
            rhom = rho * Na
            fo = association_check(self, rhom, temp_aux, Xass)
        else:
            fo = 0.

        return fo

    def ares(self, rho, T, Xass0=None):
        """
        ares(x, rho, T, Xass0)
        Method that computes the residual Helmholtz free energy of the fluid.

        Parameters
        ----------
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: float
           residual dimentionless Helmholtz free energy [Adim]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = ares(self, rho, temp_aux, Xass0)
        return a, Xass

    def dares_drho(self, rho, T, Xass0=None):
        """
        dares_drho(rho, T, Xass0)
        Method that computes the residual Helmholtz free energy of the fluid
        and its first density derivative.

        Parameters
        ----------
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           residual dimentionless Helmholtz free energy [Adim, m^3]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = dares_drho(self, rho, temp_aux, Xass0)
        return a, Xass

    def d2ares_drho(self, rho, T, Xass0=None):
        """
        d2ares_drho(rho, T, Xass0)
        Method that computes the residual Helmholtz free energy of the fluid
        and its first and second density derivatives.

        Parameters
        ----------
        rho: float
            molecular density [molecules/m3]
        T: float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array_like
           residual dimentionless Helmholtz free energy [Adim, m^3, m^6]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = d2ares_drho(self, rho, temp_aux, Xass0)
        return a, Xass

    def afcn_aux(self, rho, temp_aux, Xass0=None):
        """
        afcn_aux(rho, temp_aux, Xass0)
        Method that computes the total Helmholtz free energy of the fluid.

        Parameters
        ----------
        rho: float
            molecular density [molecules/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: float
           Helmholtz free energy [J/mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        beta = temp_aux[0]
        a, Xass = ares(self, rho, temp_aux, Xass0)
        a += aideal(rho, beta)
        a *= (Na/beta)
        return a, Xass

    def dafcn_aux(self, rho, temp_aux, Xass0=None):
        """
        dafcn_aux(rho, temp_aux, Xass0)
        Method that computes the total Helmholtz free energy of the fluid and
        its first density derivative.

        Parameters
        ----------
        rho: float
            density [mol/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array
           Helmholtz free energy and its derivative  [J/mol, J m^3/mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        beta = temp_aux[0]
        a, Xass = dares_drho(self, rho, temp_aux, Xass0)
        a += daideal_drho(rho, beta)
        a *= (Na/beta)
        return a, Xass

    def d2afcn_aux(self, rho, temp_aux, Xass0=None):
        """
        d2afcn_aux(rho, temp_aux, Xass0)
        Method that computes the total Helmholtz free energy of the fluid and
        its first ans second density derivative.

        Parameters
        ----------
        rho: float
            molecular density [molecules/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array
           Helmholtz free energy and its derivatives: a, da, d2a
           [J/mol, J m^3/mol^2,  J m^6/mol^3]
        Xass : array
            computed fraction of nonbonded sites
        """
        beta = temp_aux[0]
        a, Xass = d2ares_drho(self, rho, temp_aux, Xass0)
        a += d2aideal_drho(rho, beta)
        a *= (Na/beta)
        return a, Xass

    def afcn(self, rho, T, Xass0=None):
        """
        afcn(rho, T, Xass0)
        Method that computes the total Helmholtz free energy of the fluid.

        Parameters
        ----------
        rho: float
            molecular density [molecules/m3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: float
           Helmholtz free energy [J/mol]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = self.afcn_aux(rho, temp_aux, Xass0)
        return a

    def dafcn_drho(self, rho, T, Xass0=None):
        """
        dafcn_drho(rho, T, Xass0)
        Method that computes the total Helmholtz free energy of the fluid and
        its first density derivative.

        Parameters
        ----------
        rho: float
            molecular density [molecules/m3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array
           Helmholtz free energy and its derivative  [J/mol, J m^3/mol]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = self.dafcn_aux(rho, temp_aux, Xass0)
        return a

    def d2afcn_drho(self, rho, T, Xass0=None):
        """
        d2afcn_drho(rho, T, Xass0)
        Method that computes the total Helmholtz free energy of the fluid and
        its first ans second density derivative.

        Parameters
        ----------
        rho: float
            molecular density [molecules/m3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a: array
           Helmholtz free energy and its derivatives: a, da, d2a
           [J/mol, J m^3/mol,  J m^6/mol]
        """
        temp_aux = self.temperature_aux(T)
        a, Xass = self.d2afcn_aux(rho, temp_aux, Xass0)
        return a

    def pressure_aux(self, rho, temp_aux, Xass0=None):
        """
        pressure_aux(rho, temp_aux, Xass0)

        Method that computes the pressure at given density [mol/m3] and
        temperature [K]

        Parameters
        ----------
        rho: float
            density [mol/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        P : float
            pressure [Pa]
        Xass : array
            computed fraction of nonbonded sites
        """
        rhomolecular = Na * rho
        da, Xass = self.dafcn_aux(rhomolecular, temp_aux, Xass0)
        afcn, dafcn = da
        Psaft = rhomolecular**2 * dafcn / Na
        return Psaft, Xass

    def dP_drho_aux(self, rho, temp_aux, Xass0=None):
        """
        dP_drho_aux(rho, temp_aux, Xass0)

        Method that computes the pressure and its density derivative at given
        density [mol/m3] and temperature [K]

        Parameters
        ----------
        rho: float
            density [mol/m3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        P : float
            pressure [Pa]
        dP: float
            derivate of pressure respect density [Pa m^3 / mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        rhomolecular = Na * rho
        da, Xass = self.d2afcn_aux(rhomolecular, temp_aux, Xass0)
        afcn, dafcn, d2afcn = da
        Psaft = rhomolecular**2 * dafcn / Na
        dPsaft = 2 * rhomolecular * dafcn + rhomolecular**2 * d2afcn
        return Psaft, dPsaft, Xass

    def pressure(self, rho, T, Xass0=None):
        """
        pressure(rho, T, Xass0)

        Method that computes the pressure at given density [mol/m3] and
        temperature [K]

        Parameters
        ----------
        rho: float
            density [mol/m3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        P : float
            pressure [Pa]
        """
        temp_aux = self.temperature_aux(T)
        Psaft, Xass = self.pressure_aux(rho, temp_aux, Xass0)
        return Psaft

    def dP_drho(self, rho, T, Xass0=None):
        """
        dP_drho(rho, T, Xass0)

        Method that computes the pressure and its density derivative at given
        density [mol/m3] and temperature [K]

        Parameters
        ----------
        rho: float
            density [mol/m3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        P : float
            pressure [Pa]
        dP: float
            derivate of pressure respect density [Pa m^3 / mol]
        """
        temp_aux = self.temperature_aux(T)
        Psaft, dPsaft, Xass = self.dP_drho_aux(rho, temp_aux, Xass0)

        return Psaft, dPsaft

    def density_aux(self, temp_aux, P, state, rho0=None, Xass0=None):
        """
        density_aux(T, temp_aux, state, rho0, Xass0)
        Method that computes the density of the fluid at T, P

        Parameters
        ----------
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapor phase
        rho0 : float, optional
            initial guess to compute density root [mol/m^3]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        density: float
            density [mol/m^3]
        Xass : array
            computed fraction of nonbonded sites

        """
        if rho0 is None:
            rho, Xass = density_topliss(state, temp_aux, P, Xass0, self)
        else:
            rho, Xass = density_newton(rho0, temp_aux, P, Xass0, self)
        return rho, Xass

    def density(self, T, P, state, rho0=None, Xass0=None):
        """
        density(T, P, state)
        Method that computes the density of the fluid at T, P

        Parameters
        ----------

        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapor phase
        rho0 : float, optional
            initial guess to compute density root [mol/m^3]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        density: float
            density [mol/m^3]
        """
        temp_aux = self.temperature_aux(T)
        rho, Xass = self.density_aux(temp_aux, P, state, rho0, Xass0)
        return rho

    def psat(self, T, P0=None, v0=[None, None], Xass0=[None, None],
             full_output=False):
        """
        psat(T, P0)

        Method that computes saturation pressure at fixed T

        Parameters
        ----------

        T : float
            absolute temperature [K]
        P0 : float, optional
            initial value to find saturation pressure [Pa]
        v0: list, optional
            initial guess for liquid and vapor phase, respectively [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        full_output: bool, optional
            whether to outputs or not all the calculation info.

        Returns
        -------
        psat : float
            saturation pressure [Pa]
        vl : float
            liquid saturation volume [m3/mol]
        vv : float
            vapor saturation volume [m3/mol]
        """
        out = psat(self, T, P0, v0, Xass0, full_output)
        return out

    def tsat(self, P,  T0=None, Tbounds=None, v0=[None, None],
             Xass0=[None, None], full_output=False):
        """
        tsat(P, Tbounds)

        Method that computes saturation temperature at given pressure.

        Parameters
        ----------

        P : float
            absolute pressure [Pa]
        T0 : float, optional
             Temperature to start iterations [K]
        Tbounds : tuple, optional
                (Tmin, Tmax) Temperature interval to start iterations [K]
        v0: list, optional
            initial guess for liquid and vapor phase, respectively [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        full_output: bool, optional
            whether to outputs or not all the calculation info.

        Returns
        -------
        tsat : float
            saturation temperature [K]
        vl : float
            liquid saturation volume [m^3/mol]
        vv : float
            vapor saturation volume [m^3/mol]
        """
        out = tsat(self, P, T0, Tbounds, v0, Xass0, full_output)
        return out

    def get_critical(self, Tc0=None, rhoc0=None, method='hybr',
                     full_output=False, overwrite=False):
        """
        get_critical(Tc0, rhoc0, method)

        Method that solves the critical coordinate of the fluid.
        This metho requires good initial guesses for the critical temperature
        and density to converge.

        Second derivative of pressure against volume is estimated numerically.

        Parameters
        ----------
        Tc0 : float
            initial guess for critical temperature [K]
        rhoc : float
            initial guess for critical density [mol/m^3]
        method : string, optional
            SciPy; root method to solve critical coordinate
        full_output: bool, optional
            whether to outputs or not all the calculation info
        overwrite: bool, optional
            wheter to overwrite already computed critical points

        Returns
        -------
        Tc: float
            Critical temperature [K]
        Pc: float
            Critical pressure [Pa]
        rhoc: float
            Critical density [mol/m3]
        """
        out = get_critical(self, Tc0, rhoc0, method, full_output)

        if overwrite:
            if full_output:
                if out.success:
                    self.critical = True
                    self.Tc = out.Tc
                    self.Pc = out.Pc
                    self.rhoc = out.rhoc
            else:
                Tc0 = out[0]
                rhoc0 = out[2]
                out2 = get_critical(self, Tc0, rhoc0, method, full_output=True)
                if out2.success:
                    self.critical = True
                    self.Tc = out2.Tc
                    self.Pc = out2.Pc
                    self.rhoc = out2.rhoc
        return out

    def logfug_aux(self, temp_aux, P, state, v0=None, Xass0=None):
        """
        logfug_aux(T, P, state, v0, Xass0)

        Method that computes the fugacity coefficient at given
        composition, temperature and pressure.

        Parameters
        ----------
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        logfug : float
            fugacity coefficient
        v : float
            computed volume of the phase [m^3/mol]
        Xass : array
            computed fraction of nonbonded sites
        """
        if v0 is None:
            rho, Xass = self.density_aux(temp_aux, P, state, None, Xass0)
        else:
            rho0 = 1./v0
            rho, Xass = self.density_aux(temp_aux, P, state, rho0, Xass0)
        v = 1./rho
        rhomolecular = Na * rho
        ar, Xass = ares(self, rhomolecular, temp_aux, Xass)
        beta = temp_aux[0]
        RT = Na/beta
        Z = P * v / RT
        lnphi = ar + (Z - 1.) - np.log(Z)
        return lnphi, v, Xass

    def logfug(self, T, P, state, v0=None, Xass0=None):
        """
        logfug(T, P, state, v0, Xass0)

        Method that computes the fugacity coefficient at given temperature
        and pressure.

        Parameters
        ----------
        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        logfug: float
            fugacity coefficient
        v: float
            computed volume of the phase [m^3/mol]
        """
        temp_aux = self.temperature_aux(T)
        lnphi, v, Xass = self.logfug_aux(temp_aux, P, state, v0, Xass0)
        return lnphi, v

    def ci(self, T):
        '''
        ci(T)

        Method that evaluates the polynomial for the influence parameters used
        in the SGT theory for surface tension calculations.

        Parameters
        ----------
        T : float
            absolute temperature [K]

        Returns
        -------
        ci: float
            influence parameters [J m5 mol-2]
        '''

        return np.polyval(self.cii, T)

    def sgt_adim(self, T):
        '''
        sgt_adim(T)

        Method that evaluates dimensionless factor for temperature, pressure,
        density, tension and distance for interfacial properties computations
        with SGT.

        Parameters
        ----------
        T : float
        absolute temperature [K]

        Returns
        -------
        Tfactor : float
            factor to obtain dimentionless temperature (K -> K)
        Pfactor : float
            factor to obtain dimentionless pressure (Pa -> Pa/RT)
        rofactor : float
            factor to obtain dimentionless density (mol/m3 -> mol/m3)
        tenfactor : float
            factor to obtain surface tension in (mN/m)
        zfactor : float
            factor to obtain the interfacial distance in (Amstrongs)
        '''
        beta = 1 / (kb*T)
        RT = (Na/beta)
        cii = self.ci(T)  # computing temperature dependent cii

        Tfactor = 1.
        Pfactor = 1. / RT
        rofactor = 1.
        tenfactor = np.sqrt(cii*RT) * 1000  # To give tension in mN/m
        zfactor = 10**-10 * np.sqrt(RT / cii)

        return Tfactor, Pfactor, rofactor, tenfactor, zfactor

    def sgt_adim_fit(self, T):

        beta = 1 / (kb*T)
        RT = (Na/beta)
        Tfactor = 1.
        Pfactor = 1. / RT
        rofactor = 1.
        tenfactor = np.sqrt(RT) * 1000  # To give tension in mN/m

        return Tfactor, Pfactor, rofactor, tenfactor

    def a0ad_aux(self, rho, temp_aux, Xass0=None):
        """
        a0ad_aux(rho, temp_aux, Xass0)

        Method that computes the dimensionless Helmholtz density energy at
        given density and temperature.

        Parameters
        ----------

        rho : float
            density [mol/m^3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a0ad: float
            Helmholtz density energy divided by RT [mol/m^3]
        Xass : array
            computed fraction of nonbonded sites
        """
        rhomolecular = rho * Na
        a0, Xass = self.afcn_aux(rhomolecular, temp_aux, Xass0)
        a0 *= rho

        beta = temp_aux[0]
        RT = (Na/beta)
        a0 /= RT

        return a0, Xass

    def a0ad(self, rho, T, Xass0=None):
        """
        a0ad(rho, T, Xass0)

        Method that computes the dimensionless Helmholtz density energy at
        given density and temperature.

        Parameters
        ----------

        rho : float
            density [mol/m^3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        a0ad: float
            Helmholtz density energy divided by RT [mol/m^3]
        """
        temp_aux = self.temperature_aux(T)
        a0, Xass = self.a0ad_aux(rho, temp_aux, Xass0)
        return a0

    def muad_aux(self, rho, temp_aux, Xass0=None):
        """
        muad_aux(rho, temp_aux, Xass0)

        Method that computes the dimensionless chemical potential at given
        density and temperature.

        Parameters
        ----------
        rho : float
            density [mol/m^3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        muad: float
            chemical potential divided by RT [Adim]
        Xass : array
            computed fraction of nonbonded sites
        """

        rhomolecular = rho * Na
        da, Xass = self.dafcn_aux(rhomolecular, temp_aux, Xass0)
        afcn, dafcn = da
        mu = afcn + rhomolecular * dafcn

        beta = temp_aux[0]
        RT = (Na/beta)
        mu /= RT

        return mu, Xass

    def muad(self, rho, T, Xass0=None):
        """
        muad(rho, T, Xass0)

        Method that computes the dimensionless chemical potential at given
        density and temperature.

        Parameters
        ----------
        rho : float
            density [mol/m^3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        muad: float
            chemical potential divided by RT [Adim]
        """
        temp_aux = self.temperature_aux(T)
        mu, Xass = self.muad_aux(rho, temp_aux, Xass0)
        return mu

    def dOm_aux(self, rho, temp_aux, mu, Psat, Xass0=None):
        """
        dOm_aux(rho, temp_aux, mu, Psat, Xass0)

        Method that computes the dimensionless Thermodynamic Grand potential
        at given density and temperature.

        Parameters
        ----------
        rho : float
            density [mol/m^3]
        temp_aux : list
            temperature dependend parameters computed with temperature_aux(T)
        mu : float
            dimensionless chemical potential at equilibrium [Adim]
        Psat : float
            equilibrium pressure divided by RT [Pa mol / J]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        GPT: float
            Thermodynamic Grand potential divided by RT [Pa mol / J]
        Xass : array
            computed fraction of nonbonded sites
        """
        a0, Xass = self.a0ad_aux(rho, temp_aux, Xass0)
        GPT = a0 - rho*mu + Psat

        return GPT, Xass

    def dOm(self, rho, T, mu, Psat, Xass0=None):
        """
        dOm(rho, T, mu, Psat, Xass0)

        Method that computes the dimensionless Thermodynamic Grand potential
        at given density and temperature.

        Parameters
        ----------
        rho : float
            density [mol/m^3]
        T : float
            absolute temperature [K]
        mu : float
            dimensionless chemical potential at equilibrium [Adim]
        Psat : float
            equilibrium pressure divided by RT [Pa mol / J]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites

        Returns
        -------
        Out: float
            Thermodynamic Grand potential divided by RT [Pa mol / J]
        """
        temp_aux = self.temperature_aux(T)
        GPT, Xass = self.dOm_aux(rho, temp_aux, mu, Psat, Xass0)
        return GPT

    def EntropyR(self, T, P, state, v0=None, Xass0=None, T_step=0.1):
        """
        EntropyR(T, P, state, v0, Xass0, T_step)

        Method that computes the residual entropy at given temperature and
        pressure.

        Parameters
        ----------
        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        T_step: float, optional
            Step to compute the numerical temperature derivates of Helmholtz
            free energy

        Returns
        -------
        Sr : float
            residual entropy [J/mol K]

        """
        temp_aux = self.temperature_aux(T)
        if v0 is None:
            rho, Xass = self.density_aux(temp_aux, P, state, None, Xass0)
        else:
            rho0 = 1./v0
            rho, Xass = self.density_aux(temp_aux, P, state, rho0, Xass0)
        v = 1./rho
        rhomolecular = Na * rho

        a, Xass = ares(self, rhomolecular, temp_aux, Xass)
        beta = temp_aux[0]
        RT = Na/beta
        Z = P * v / RT

        h = T_step
        temp_aux1 = self.temperature_aux(T+h)
        temp_aux2 = self.temperature_aux(T+2*h)
        temp_aux_1 = self.temperature_aux(T-h)
        temp_aux_2 = self.temperature_aux(T-2*h)

        a1, Xass1 = ares(self, rhomolecular, temp_aux1, Xass)
        a2, Xass2 = ares(self, rhomolecular, temp_aux2, Xass)
        a_1, Xass_1 = ares(self, rhomolecular, temp_aux_1, Xass)
        a_2, Xass_2 = ares(self, rhomolecular, temp_aux_2, Xass)

        F = a
        dFdT = (a_2/12 - 2*a_1/3 + 2*a1/3 - a2/12)/h
        Sr_TVN = -T*dFdT - F  # residual entropy (TVN) divided by R
        Sr_TPN = Sr_TVN + np.log(Z)  # residual entropy (TPN) divided by R
        Sr_TPN *= R
        return Sr_TPN

    def EnthalpyR(self, T, P, state, v0=None, Xass0=None, T_step=0.1):
        """
        EnthalpyR(T, P, state, v0, Xass0, T_step)

        Method that computes the residual enthalpy at given temperature and
        pressure.

        Parameters
        ----------
        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        T_step: float, optional
            Step to compute the numerical temperature derivates of Helmholtz
            free energy

        Returns
        -------
        Hr : float
            residual enthalpy [J/mol]

        """
        temp_aux = self.temperature_aux(T)
        if v0 is None:
            rho, Xass = self.density_aux(temp_aux, P, state, None, Xass0)
        else:
            rho0 = 1./v0
            rho, Xass = self.density_aux(temp_aux, P, state, rho0, Xass0)
        v = 1./rho
        rhomolecular = Na * rho

        a, Xass = ares(self, rhomolecular, temp_aux, Xass)
        beta = temp_aux[0]
        RT = Na/beta
        Z = P * v / RT

        h = T_step
        temp_aux1 = self.temperature_aux(T+h)
        temp_aux2 = self.temperature_aux(T+2*h)
        temp_aux_1 = self.temperature_aux(T-h)
        temp_aux_2 = self.temperature_aux(T-2*h)

        a1, Xass1 = ares(self, rhomolecular, temp_aux1, Xass)
        a2, Xass2 = ares(self, rhomolecular, temp_aux2, Xass)
        a_1, Xass_1 = ares(self, rhomolecular, temp_aux_1, Xass)
        a_2, Xass_2 = ares(self, rhomolecular, temp_aux_2, Xass)

        F = a
        dFdT = (a_2/12 - 2*a_1/3 + 2*a1/3 - a2/12)/h
        Sr_TVN = -T*dFdT - F  # residual entropy divided by R
        Hr_TPN = F + Sr_TVN + Z - 1.  # residual entalphy divided by RT
        Hr_TPN *= RT
        return Hr_TPN

    def CvR(self, rho, T, Xass0=None, T_step=0.1):
        """
        CvR(rho, T, Xass0, T_step)

        Method that computes the residual isochoric heat capacity at given
        density and temperature.

        Parameters
        ----------
        rho : float
            density [mol/m^3]
        T : float
            absolute temperature [K]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        T_step: float, optional
            Step to compute temperature numerical derivates of Helmholtz
            free energy

        Returns
        -------
        Cv: float
            isochoric heat capacity [J/mol K]
        """
        temp_aux = self.temperature_aux(T)

        rhomolecular = Na * rho

        a, Xass = ares(self, rhomolecular, temp_aux, Xass0)

        h = T_step
        temp_aux1 = self.temperature_aux(T+h)
        temp_aux2 = self.temperature_aux(T+2*h)
        temp_aux_1 = self.temperature_aux(T-h)
        temp_aux_2 = self.temperature_aux(T-2*h)

        a1, Xass1 = ares(self, rhomolecular, temp_aux1, Xass)
        a2, Xass2 = ares(self, rhomolecular, temp_aux2, Xass)
        a_1, Xass_1 = ares(self, rhomolecular, temp_aux_1, Xass)
        a_2, Xass_2 = ares(self, rhomolecular, temp_aux_2, Xass)

        dFdT = (a_2/12 - 2*a_1/3 + 2*a1/3 - a2/12)/h
        d2FdT = (-a_2/12 + 4*a_1/3 - 5*a/2 + 4*a1/3 - a2/12)/h**2

        Cvr_TVN = -T**2*d2FdT - 2*T*dFdT  # residual isochoric heat capacity
        Cvr_TVN *= R
        return Cvr_TVN

    def CpR(self, T, P, state, v0=None, Xass0=None, T_step=0.1):
        """
        Cpr(T, P, state, v0, Xass0, T_step)

        Method that computes the residual heat capacity at given temperature
        and pressure.

        Parameters
        ----------
        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        T_step: float, optional
            Step to compute the numerical temperature derivates of Helmholtz
            free energy

        Returns
        -------
        Cp: float
            residual heat capacity [J/mol K]
        """

        temp_aux = self.temperature_aux(T)
        if v0 is None:
            rho, Xass = self.density_aux(temp_aux, P, state, None, Xass0)
        else:
            rho0 = 1./v0
            rho, Xass = self.density_aux(temp_aux, P, state, rho0, Xass0)

        rhomolecular = Na * rho

        d2a, Xass = d2ares_drho(self, rhomolecular, temp_aux, Xass)
        beta = temp_aux[0]
        RT = Na/beta

        h = T_step
        temp_aux1 = self.temperature_aux(T+h)
        temp_aux2 = self.temperature_aux(T+2*h)
        temp_aux_1 = self.temperature_aux(T-h)
        temp_aux_2 = self.temperature_aux(T-2*h)

        a1, Xass1 = dares_drho(self, rhomolecular, temp_aux1, Xass)
        a2, Xass2 = dares_drho(self, rhomolecular, temp_aux2, Xass)
        a_1, Xass_1 = dares_drho(self, rhomolecular, temp_aux_1, Xass)
        a_2, Xass_2 = dares_drho(self, rhomolecular, temp_aux_2, Xass)

        a = d2a[:2]
        da_drho = a[1] * Na
        d2a_drho = d2a[2] * Na**2

        dFdT = (a_2/12 - 2*a_1/3 + 2*a1/3 - a2/12)/h
        dFdT[1] *= Na

        d2FdT = (-a_2/12 + 4*a_1/3 - 5*a/2 + 4*a1/3 - a2/12) / h**2
        d2FdT[1] *= Na

        dP_dT = RT*(rho**2 * dFdT[1]) + P/T

        dP_drho = 2*rho*da_drho + 2.
        dP_drho += rho**2 * d2a_drho - 1.
        dP_drho *= RT

        dP_dV = -rho**2 * dP_drho
        # residual isochoric heat capacity
        Cvr_TVN = R * (-T**2*d2FdT[0] - 2*T*dFdT[0])
        # residual heat capacity
        Cpr = Cvr_TVN - R - T*dP_dT**2/dP_dV
        return Cpr

    def speed_sound(self, T, P, state, v0=None, Xass0=None, T_step=0.1,
                    CvId=3*R/2, CpId=5*R/2):
        """
        speed_sound(T, P, state, v0, Xass0, T_step, CvId, CpId)

        Method that computes the speed of sound at given temperature
        and pressure.

        This calculation requires that the molar weight of the fluid has been
        set in the component function.

        By default the ideal gas Cv and Cp are set to 3R/2 and 5R/2, the user
        can supply better values if available.

        Parameters
        ----------
        T : float
            absolute temperature [K]
        P : float
            pressure [Pa]
        state : string
            'L' for liquid phase and 'V' for vapour phase
        v0: float, optional
            initial guess for volume root [m^3/mol]
        Xass0: array, optional
            Initial guess for the calculation of fraction of non-bonded sites
        T_step: float, optional
            Step to compute the numerical temperature derivates of Helmholtz
            free energy
        CvId: float, optional
            Ideal gas isochoric heat capacity, set to 3R/2 by default [J/mol K]
        CpId: float, optional
            Ideal gas heat capacity, set to 3R/2 by default [J/mol K]


        Returns
        -------
        w: float
            speed of sound [m/s]
        """

        temp_aux = self.temperature_aux(T)
        if v0 is None:
            rho, Xass = self.density_aux(temp_aux, P, state, None, Xass0)
        else:
            rho0 = 1./v0
            rho, Xass = self.density_aux(temp_aux, P, state, rho0, Xass0)

        rhomolecular = Na * rho

        d2a, Xass = d2ares_drho(self, rhomolecular, temp_aux, Xass)
        beta = temp_aux[0]
        RT = Na/beta

        h = T_step
        temp_aux1 = self.temperature_aux(T+h)
        temp_aux2 = self.temperature_aux(T+2*h)
        temp_aux_1 = self.temperature_aux(T-h)
        temp_aux_2 = self.temperature_aux(T-2*h)

        a1, Xass1 = dares_drho(self, rhomolecular, temp_aux1, Xass)
        a2, Xass2 = dares_drho(self, rhomolecular, temp_aux2, Xass)
        a_1, Xass_1 = dares_drho(self, rhomolecular, temp_aux_1, Xass)
        a_2, Xass_2 = dares_drho(self, rhomolecular, temp_aux_2, Xass)

        a = d2a[:2]
        da_drho = a[1] * Na
        d2a_drho = d2a[2] * Na**2

        dFdT = (a_2/12 - 2*a_1/3 + 2*a1/3 - a2/12)/h
        dFdT[1] *= Na

        d2FdT = (-a_2/12 + 4*a_1/3 - 5*a/2 + 4*a1/3 - a2/12) / h**2
        d2FdT[1] *= Na

        dP_dT = RT*(rho**2 * dFdT[1]) + P/T

        dP_drho = 2*rho*da_drho + 2.
        dP_drho += rho**2 * d2a_drho - 1.
        dP_drho *= RT
        dP_dV = -rho**2 * dP_drho

        # residual isochoric heat capacity
        Cvr_TVN = R * (-T**2*d2FdT[0] - 2*T*dFdT[0])
        # residual heat capacity
        Cpr = Cvr_TVN - R - T*dP_dT**2/dP_dV

        # speed of sound calculation
        Cp = CpId + Cpr
        Cv = CvId + Cvr_TVN

        betas = -rho * (Cv/Cp) / dP_dV

        w2 = 1000./(rho * betas * self.Mw)
        w = np.sqrt(w2)

        return w



from __future__ import division, print_function, absolute_import

# Constants
kb = 1.3806488e-23  # [J/K] Boltzman's constant
Na = 6.022142e23  # [mol-1] Avogadro's Number
R = Na * kb    # [J mol-1 K-1] Ideal gas constant



from __future__ import division, print_function, absolute_import
from .vrmie_mixtures.saftvrmiemix import saftvrmie_mix
from .vrmie_pure.saftvrmie import saftvrmie_pure
from .gammamie_mixtures.saftgammamie_mixture import saftgammamie_mix
from .gammamie_pure.saftgammamie_pure import saftgammamie_pure


def saftvrmie(mix_or_component, compute_critical=True):
    '''
    Returns SAFT-VR-Mie EoS object.

    Parameters
    ----------
    mix_or_component : object
        :class:`SGTPy.mixture` or :class:`SGTPy.component` object
    compute_critical: bool
        If True the critical point of the fluid will attempt to be computed
        (it might fail for some fluids).
    Returns
    -------
    eos : object
        SAFT-VR-Mie EoS object
    '''
    nc = mix_or_component.nc
    if nc == 1:
        eos = saftvrmie_pure(mix_or_component, compute_critical)
    else:
        eos = saftvrmie_mix(mix_or_component, compute_critical)
    return eos


def saftgammamie(mix_or_component, compute_critical=True):
    '''
    Returns SAFT-Gamma-Mie EoS object.

    Parameters
    ----------
    mix_or_component : object
        :class:`saftgammamie.mixture` or :class:`saftgammamie.component` object
    compute_critical: bool
        If True the critical point of the fluid will attempt to be computed
        (it might fail for some fluids).
    Returns
    -------
    eos : object
        SAFT-Gamma-Mie EoS object
    '''
    nc = mix_or_component.nc
    if nc == 1:
        eos = saftgammamie_pure(mix_or_component, compute_critical)
    else:
        eos = saftgammamie_mix(mix_or_component, compute_critical)
    return eos



from __future__ import division, print_function, absolute_import
import numpy as np


# Equation A29 to A33
def kichain(xhix):
    xhix2 = xhix**2
    xhix3 = xhix**3
    xhix4 = xhix2**2
    xhix_1 = (1 - xhix)
    xhix_13 = xhix_1**3
    k0 = -np.log(xhix_1) + (42*xhix - 39*xhix2 + 9*xhix3 - 2*xhix4)/(6*xhix_13)
    k1 = (xhix4 + 6*xhix2 - 12*xhix)/(2*xhix_13)
    k2 = -3*xhix2/(8*xhix_1**2)
    k3 = (-xhix4 + 3*xhix2 + 3*xhix)/(6*xhix_13)
    return np.array([k0, k1, k2, k3])


def dkichain_dxhi00(xhix, dxhix_dxhi00):

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_1 = (1 - xhix)
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1

    k0 = -np.log(xhix_1) + (42*xhix - 39*xhix2 + 9*xhix3 - 2*xhix4)/(6*xhix_13)
    k1 = (xhix4 + 6*xhix2 - 12*xhix)/(2*xhix_13)
    k2 = -3*xhix2/(8*xhix_1**2)
    k3 = (-xhix4 + 3*xhix2 + 3*xhix)/(6*xhix_13)

    dk0 = 24 + xhix * (-6 + xhix * (3 - 7 * xhix + xhix2))
    dk0 *= dxhix_dxhi00 / 3. / xhix_14

    dk1 = 12 + xhix * (2+xhix) * (6 - 6 * xhix + xhix2)
    dk1 *= -dxhix_dxhi00 / 2. / xhix_14

    dk2 = 3*xhix
    dk2 *= dxhix_dxhi00 / 4. / -xhix_13

    dk3 = 3 + xhix * (12 + (-3 + xhix) * (-xhix + xhix2))
    dk3 *= dxhix_dxhi00 / 6. / xhix_14

    dks = np.array([[k0, k1, k2, k3], [dk0, dk1, dk2, dk3]])
    return dks


def d2kichain_dxhi00(xhix, dxhix_dxhi00):

    dxhix_dxhi00_2 = dxhix_dxhi00**2

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_1 = (1 - xhix)
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1
    xhix_15 = xhix_14*xhix_1

    k0 = -np.log(xhix_1) + (42*xhix - 39*xhix2 + 9*xhix3 - 2*xhix4)/(6*xhix_13)
    k1 = (xhix4 + 6*xhix2 - 12*xhix)/(2*xhix_13)
    k2 = -3*xhix2/(8*xhix_1**2)
    k3 = (-xhix4 + 3*xhix2 + 3*xhix)/(6*xhix_13)

    dk0 = 24 + xhix * (-6 + xhix * (3 - 7 * xhix + xhix2))
    dk0 *= dxhix_dxhi00 / 3. / xhix_14

    dk1 = 12 + xhix * (2+xhix) * (6 - 6 * xhix + xhix2)
    dk1 *= -dxhix_dxhi00 / 2. / xhix_14

    dk2 = 3*xhix
    dk2 *= dxhix_dxhi00 / 4. / -xhix_13

    dk3 = 3 + xhix * (12 + (-3 + xhix) * (-xhix + xhix2))
    dk3 *= dxhix_dxhi00 / 6. / xhix_14

    d2k0 = 3 * (-30 + xhix * (1+xhix) * (4 + xhix))
    d2k0 *= dxhix_dxhi00_2 / 3. / -xhix_15

    d2k1 = 12 * (5 - 2 * (-1 + xhix) * xhix)
    d2k1 *= dxhix_dxhi00_2 / 2. / -xhix_15

    d2k2 = -3*(1+2*xhix)
    d2k2 *= dxhix_dxhi00_2 / 4. / xhix_14

    d2k3 = 6*(-4 + xhix * (-7 + xhix))
    d2k3 *= dxhix_dxhi00_2 / 6. / -xhix_15

    d2ks = np.array([[k0, k1, k2, k3],
                     [dk0, dk1, dk2, dk3],
                     [d2k0, d2k1, d2k2, d2k3]])
    return d2ks


def dkichain_dx(xhix, dxhix_dx):

    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_1 = (1 - xhix)
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1

    k0 = -np.log(xhix_1) + (42*xhix - 39*xhix2 + 9*xhix3 - 2*xhix4)/(6*xhix_13)
    k1 = (xhix4 + 6*xhix2 - 12*xhix)/(2*xhix_13)
    k2 = -3*xhix2/(8*xhix_1**2)
    k3 = (-xhix4 + 3*xhix2 + 3*xhix)/(6*xhix_13)

    dk0 = 24 + xhix * (-6 + xhix * (3 - 7 * xhix + xhix2))
    dk0 *= dxhix_dx / 3. / xhix_14

    dk1 = 12 + xhix * (2+xhix) * (6 - 6 * xhix + xhix2)
    dk1 *= -dxhix_dx / 2. / xhix_14

    dk2 = 3*xhix
    dk2 *= dxhix_dx / 4. / -xhix_13

    dk3 = 3 + xhix * (12 + (-3 + xhix) * (-xhix + xhix2))
    dk3 *= dxhix_dx / 6. / xhix_14

    ks = np.array([k0, k1, k2, k3])
    dksx = np.array([dk0, dk1, dk2, dk3]).T
    return ks, dksx


def dkichain_dx_dxhi00(xhix, dxhix_dxhi00, dxhix_dx):
    xhix2 = xhix**2
    xhix3 = xhix2*xhix
    xhix4 = xhix3*xhix
    xhix_1 = (1 - xhix)
    xhix_13 = xhix_1**3
    xhix_14 = xhix_13*xhix_1

    k0 = -np.log(xhix_1) + (42*xhix - 39*xhix2 + 9*xhix3 - 2*xhix4)/(6*xhix_13)
    k1 = (xhix4 + 6*xhix2 - 12*xhix)/(2*xhix_13)
    k2 = -3*xhix2/(8*xhix_1**2)
    k3 = (-xhix4 + 3*xhix2 + 3*xhix)/(6*xhix_13)

    aux_k0 = 24 + xhix * (-6 + xhix * (3 - 7 * xhix + xhix2))
    aux_k0 /= 3.*xhix_14
    aux_k1 = 12 + xhix * (2+xhix) * (6 - 6 * xhix + xhix2)
    aux_k1 /= -2.*xhix_14
    aux_k2 = 3*xhix / 4. / -xhix_13
    aux_k3 = 3 + xhix * (12 + (-3 + xhix) * (-xhix + xhix2))
    aux_k3 /= 6.*xhix_14

    dks = np.array([[k0, k1, k2, k3], [aux_k0, aux_k1, aux_k2, aux_k3]])
    dksx = np.multiply.outer(dxhix_dx, dks[1])

    dks[1] *= dxhix_dxhi00

    return dks, dksx


def gdHS(x0i_matrix, xhix):
    ks = kichain(xhix)
    # x0i_matrix = np.array([x0i**0, x0i, x0i**2, x0i**3])
    g = np.exp(np.dot(ks, x0i_matrix))
    return g


def dgdHS_dxhi00(x0i_matrix, xhix, dxhix_dxhi00):
    dks = dkichain_dxhi00(xhix, dxhix_dxhi00)
    # x0i_matrix = np.array([x0i**0, x0i, x0i**2, x0i**3])
    dg = np.dot(dks, x0i_matrix)
    dg[0] = np.exp(dg[0])
    dg[1] *= dg[0]
    return dg


def d2gdHS_dxhi00(x0i_matrix, xhix, dxhix_dxhi00):
    d2ks = d2kichain_dxhi00(xhix, dxhix_dxhi00)
    # x0i_matrix = np.array([x0i**0, x0i, x0i**2, x0i**3])
    d2g = np.matmul(d2ks, x0i_matrix)
    d2g[0] = np.exp(d2g[0])
    d2g[2] += d2g[1]**2
    d2g[2] *= d2g[0]
    d2g[1] *= d2g[0]
    return d2g


def dgdHS_dx(x0i_matrix, xhix, dxhix_dx):
    ks, dksx = dkichain_dx(xhix, dxhix_dx)
    # x0i_matrix = np.array([x0i**0, x0i, x0i**2, x0i**3])
    g = g = np.exp(np.dot(ks, x0i_matrix))
    dgx = g * np.matmul(dksx, x0i_matrix)
    return g, dgx


def dgdHS_dxxhi(x0i_matrix, xhix, dxhix_dxhi00, dxhix_dx):
    dks, dksx = dkichain_dx_dxhi00(xhix, dxhix_dxhi00, dxhix_dx)
    dg = np.dot(dks, x0i_matrix)
    dg[0] = np.exp(dg[0])
    dg[1] *= dg[0]
    # x0i_matrix = np.array([x0i**0, x0i, x0i**2, x0i**3])
    dgx = dg[0] * np.matmul(dksx, x0i_matrix)
    return dg, dgx



from __future__ import division, print_function, absolute_import
import numpy as np
from .monomer_aux import Xi, dXi_dxhi00, d2Xi_dxhi00, dXi_dx, dXi_dxxhi


def a2(xs, khs, xhixm, a2ij, epsij, f1, f2, f3):

    xi = Xi(xhixm, f1, f2, f3)
    a2 = a2ij * khs * (1 + xi) * epsij/2
    a = np.matmul(np.matmul(a2, xs), xs)

    return a


def da2_dxhi00(xs, khs, dkhs, xhixm, dxhixm_dxhi00, da2ij, epsij, f1, f2, f3):

    xi, dxi = dXi_dxhi00(xhixm, dxhixm_dxhi00, f1, f2, f3)
    xi1 = 1 + xi
    ctes = epsij / 2
    sum1, dsum1 = da2ij

    a2 = sum1 * khs * xi1 * ctes
    da2 = sum1 * xi1 * dkhs + khs * xi1 * dsum1 + khs * sum1 * dxi
    da2 *= ctes

    da = np.array([a2, da2])
    a = np.matmul(np.matmul(da, xs), xs)
    return a


def d2a2_dxhi00(xs, khs, dkhs, d2khs, xhixm, dxhixm_dxhi00, d2a2ij, epsij,
                f1, f2, f3):

    xi, dxi, d2xi = d2Xi_dxhi00(xhixm, dxhixm_dxhi00, f1, f2, f3)
    xi1 = 1 + xi
    ctes = epsij / 2
    sum1, dsum1, d2sum1 = d2a2ij

    a2 = sum1 * khs * xi1 * ctes

    da2 = sum1 * xi1 * dkhs + khs * xi1 * dsum1 + khs * sum1 * dxi
    da2 *= ctes

    d2a2 = sum1 * xi1 * d2khs
    d2a2 += sum1 * khs * d2xi
    d2a2 += d2sum1 * xi1 * khs
    d2a2 += 2 * dkhs * (xi1 * dsum1 + sum1 * dxi)
    d2a2 += 2 * khs * dsum1 * dxi
    d2a2 *= ctes

    da = np.array([a2, da2, d2a2])
    a = np.matmul(np.matmul(da, xs), xs)
    return a


def da2_dx(xs, dxs_dx, khs, dkhsx, xhixm, dxhixm_dx, a2ij, da2ijx, epsij, f1,
           f2, f3):

    xi, dxi = dXi_dx(xhixm, dxhixm_dx, f1, f2, f3)
    xi1 = 1 + xi
    ctes = epsij/2
    aux0 = xi1 * a2ij
    aux1 = aux0 * khs
    aux2 = aux1 * ctes

    a = np.dot(xs, np.dot(aux2, xs))

    da2 = np.multiply.outer(dkhsx, aux0)
    da2 += a2ij * khs * dxi
    da2 += khs * xi1 * da2ijx
    da2 *= ctes

    aux3 = aux2 * xs
    suma1 = 2*np.sum(dxs_dx@aux3, axis=1)
    dax = np.matmul(np.matmul(xs, da2), xs) + suma1

    return a, dax


def da2_dxxhi(xs, dxs_dx, khs, dkhs, dkhsx, xhixm, dxhixm_dxhi00, dxhixm_dx,
              da2ij, da2ijx, epsij, f1, f2, f3):

    ctes = epsij/2

    xi, dxi, dxix = dXi_dxxhi(xhixm, dxhixm_dxhi00, dxhixm_dx, f1, f2, f3)
    xi1 = 1 + xi

    sum1, dsum1 = da2ij

    aux0 = xi1 * sum1
    aux1 = xi1 * khs
    aux2 = sum1 * khs

    aux3 = aux0 * khs
    aux4 = aux3 * ctes

    a2 = aux4
    da2 = aux0 * dkhs + aux1 * dsum1 + aux2 * dxi
    da2 *= ctes

    da = np.array([a2, da2])
    a = np.matmul(np.matmul(xs, da), xs)

    da2 = np.multiply.outer(dkhsx, aux0)
    da2 += aux1 * da2ijx
    da2 += aux2 * dxix
    da2 *= ctes

    aux5 = xs * aux4
    suma1 = 2*np.sum(dxs_dx@aux5, axis=1)
    dax = np.matmul(np.matmul(xs, da2), xs) + suma1

    return a, dax



from __future__ import division, print_function, absolute_import
import numpy as np
# from .monomer_aux import I_lam, J_lam


def B(eta, Ilam, Jlam, eps):
    # B calculation Eq 33
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)
    eta13 = (1-eta)**3
    ter1 = (1-eta/2)*Ilam/eta13 - 9.*eta*(1+eta)*Jlam/(2.*eta13)
    b = 12.*eta*eps*ter1
    return b


def dB(eta, Ilam, Jlam, eps):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)

    eta13 = (1-eta)**3.
    ter1 = (1.-eta/2.)*Ilam/eta13 - 9.*eta*(1+eta)*Jlam/(2.*eta13)
    b = 12.*eta*eps*ter1

    # first derivative
    ter1b1 = Ilam*(-2. + (eta-2.)*eta) + Jlam*18.*eta*(1.+2.*eta)
    ter2b1 = -6.*eps/(eta-1.)**4.
    db = ter1b1*ter2b1
    return np.hstack([b, db])


def d2B(eta, Ilam, Jlam, eps):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)

    eta13 = (1.-eta)**3.
    ter1 = (1.-eta/2.)*Ilam/eta13 - 9.*eta*(1.+eta)*Jlam/(2.*eta13)
    b = 12.*eta*eps*ter1

    # first derivative
    db = Ilam*(-2. + (eta-2)*eta) + Jlam*18.*eta*(1.+2.*eta)
    db *= (-6.*eps/(eta-1.)**4)

    # second derivative
    d2b = Ilam*(-5. + (eta-2)*eta) + Jlam*9.*(1 + eta*(7. + 4.*eta))
    d2b *= 12.*eps/(eta-1)**5
    return np.hstack([b, db, d2b])


def d3B(eta, Ilam, Jlam, eps):
    # I = I_lam(x0, lam)
    # J = J_lam(x0, lam)

    eta13 = (1.-eta)**3.
    ter1 = (1.-eta/2.)*Ilam/eta13 - 9.*eta*(1.+eta)*Jlam/(2.*eta13)
    b = 12.*eta*eps*ter1

    # first derivative
    db = Ilam*(-2. + (eta-2.)*eta) + Jlam*18.*eta*(1.+2.*eta)
    db *= (-6.*eps/(eta-1.)**4.)

    # second derivative
    d2b = Ilam*(-5. + (eta-2.)*eta) + Jlam*9.*(1. + eta*(7.+4.*eta))
    d2b *= 12.*eps/(eta-1.)**5.

    d3b = Ilam*(9. - (eta-2.)*eta) - Jlam*36.*(1. + eta*(3. + eta))
    d3b *= (36.*eps/(1.-eta)**6.)
    return np.hstack([b, db, d2b, d3b])



from __future__ import division, print_function, absolute_import
import numpy as np
from ..constants import Na
from scipy.optimize import minimize_scalar, toms748


def dPsaft_fun(rho, temp_aux, saft):
    rhomolecular = Na * rho
    global Xass
    da, Xass = saft.d2afcn_aux(rhomolecular, temp_aux, Xass)
    afcn, dafcn, d2afcn = da
    dPsaft = 2 * rhomolecular * dafcn + rhomolecular**2 * d2afcn
    return dPsaft


def Psaft_obj(rho, temp_aux, saft, Pspec):
    rhomolecular = Na * rho
    global Xass
    da, Xass = saft.dafcn_aux(rhomolecular, temp_aux, Xass)
    afcn, dafcn = da
    Psaft = rhomolecular**2 * dafcn / Na
    return Psaft - Pspec


def density_newton_lim(rho_a, rho_b, temp_aux, P, Xass0, saft):
    rho = (rho_a + rho_b) / 2
    Psaft, dPsaft, Xass = saft.dP_drho_aux(rho, temp_aux, Xass0)
    for i in range(15):
        rho_old = rho
        FO = Psaft - P
        dFO = dPsaft
        drho = FO/dFO
        rho_new = rho - drho

        if FO > 0:
            rho_b = rho
        else:
            rho_a = rho

        if rho_a < rho_new < rho_b:
            rho = rho_new
        else:
            rho = (rho_a + rho_b) / 2
        if np.abs(rho - rho_old) < 1e-5:
            break
        Psaft, dPsaft, Xass = saft.dP_drho_aux(rho, temp_aux, Xass)
    return rho, Xass


def density_topliss(state, temp_aux, P, Xass0, saft):

    if state != 'L' and state != 'V':
        raise Warning("Not valid state. 'L' for liquid and 'V' for vapor.")

    beta = temp_aux[0]
    # lower boundary a zero density
    rho_lb = 1e-5
    dP_lb = Na / beta

    # upper boundary limit at infinity pressure
    etamax = 0.7405
    rho_lim = (6 * etamax) / (saft.xs_m * np.pi * saft.sigma_ii3) / Na
    ub_sucess = False
    rho_ub = 0.4 * rho_lim
    it = 0
    P_ub, dP_ub, Xass_ub = saft.dP_drho_aux(rho_ub, temp_aux, Xass0)
    while not ub_sucess and it < 5:
        it += 1
        P_ub, dP_ub, Xass_ub = saft.dP_drho_aux(rho_ub, temp_aux, Xass_ub)
        rho_ub += 0.15 * rho_lim
        ub_sucess = P_ub > P and dP_ub > 0

    # derivative computation  at zero density
    rho_lb1 = 1e-4 * rho_lim
    P_lb1, dP_lb1, Xass_lb = saft.dP_drho_aux(rho_lb1, temp_aux, Xass0)
    d2P_lb1 = (dP_lb1 - dP_lb) / rho_lb1
    if d2P_lb1 > 0:
        flag = 3
    else:
        flag = 1

    global Xass
    Xass = Xass0

    # Stage 1
    bracket = [rho_lb, rho_ub]
    if flag == 1:
        # Found inflexion point
        sol_inf = minimize_scalar(dPsaft_fun, args=(temp_aux, saft),
                                  bounds=bracket, method='Bounded',
                                  options={'xatol': 1e-2})
        rho_inf = sol_inf.x
        dP_inf = sol_inf.fun
        if dP_inf > 0:
            flag = 3
        else:
            flag = 2

    # Stage 2
    if flag == 2:
        if state == 'L':
            bracket[0] = rho_inf
        elif state == 'V':
            bracket[1] = rho_inf
        rho_ext = toms748(dPsaft_fun, bracket[0], bracket[1],
                          args=(temp_aux, saft), xtol=1e-2)
        P_ext, dP_ext, Xass = saft.dP_drho_aux(rho_ext, temp_aux, Xass)
        if P_ext > P and state == 'V':
            bracket[1] = rho_ext
        elif P_ext < P and state == 'L':
            bracket[0] = rho_ext
        else:
            flag = -1

    if flag == -1:
        rho = np.nan
    else:
        rho, Xass = density_newton_lim(bracket[0], bracket[1], temp_aux,
                                       P, Xass, saft)

    return rho, Xass


def density_newton(rho0, temp_aux, P, Xass0, saft):

    rho = 1.*rho0
    Psaft, dPsaft, Xass = saft.dP_drho_aux(rho, temp_aux, Xass0)
    for i in range(15):
        FO = Psaft - P
        dFO = dPsaft
        drho = FO/dFO
        rho -= drho
        if np.abs(drho) < 1e-5:
            break
        Psaft, dPsaft, Xass = saft.dP_drho_aux(rho, temp_aux, Xass)
    return rho, Xass



from __future__ import division, print_function, absolute_import
import numpy as np
from scipy.optimize import root
from ..math import lobatto
from .cijmix_cy import cmix_cy
from .tensionresult import TensionResult


def fobj_saddle(ros, mu0, T, eos):
    mu = eos.muad(ros, T)
    return mu - mu0


def sgt_linear(rho1, rho2, Tsat, Psat, model, n=100, full_output=False,
               check_eq=True):
    """
    SGT linear for mixtures (rho1, rho2, T, P) -> interfacial tension

    Parameters
    ----------
    rho1 : float
        phase 1 density vector
    rho2 : float
        phase 2 density vector
    Tsat : float
        saturation temperature
    Psat : float
        saturation pressure
    model : object
        created with an EoS
    n : int, optional
        number points to solve density profiles
    full_output : bool, optional
        wheter to outputs all calculation info
    check_eq : bool, optional
        whether to check if given density vectors are in phase equilibria

    Returns
    -------
    ten : float
        interfacial tension between the phases
    """

    # Dimensionless variables
    Tfactor, Pfactor, rofactor, tenfactor, zfactor = model.sgt_adim(Tsat)
    Pad = Psat*Pfactor
    ro1a = rho1*rofactor
    ro2a = rho2*rofactor

    cij = model.ci(Tsat)
    cij /= cij[0, 0]

    mu0 = model.muad(ro1a, Tsat)
    mu02 = model.muad(ro2a, Tsat)
    if check_eq:
        if not np.allclose(mu0, mu02, rtol=1e-3):
            raise Exception('Not equilibria compositions, mu1 != mu2')
    roots, weights = lobatto(n)
    s = 0
    ro_s = (ro2a[s]-ro1a[s])*roots + ro1a[s]  # integrations nodes
    wreal = np.abs(weights*(ro2a[s]-ro1a[s]))  # integration weights

    # Linear profile
    pend = (ro2a - ro1a)
    b = ro1a
    ro = (np.outer(roots, pend) + b).T

    # Derivatives respect to component 1
    dro = np.gradient(ro, ro_s, edge_order=2, axis=1)

    suma = cmix_cy(dro, cij)
    dom = np.zeros(n)
    for k in range(1, n - 1):
        dom[k] = model.dOm(ro[:, k], Tsat, mu0, Pad)

    integral = np.nan_to_num(np.sqrt(2*dom*suma))
    tension = np.dot(integral, wreal)
    tension *= tenfactor

    if full_output:
        # Z profile
        with np.errstate(divide='ignore'):
            intz = (np.sqrt(suma/(2*dom)))
        intz[np.isinf(intz)] = 0
        z = np.cumsum(intz*wreal)
        z /= zfactor
        ro /= rofactor
        dictresult = {'tension': tension, 'rho': ro, 'z': z,
                      'GPT': np.hstack([0, dom, 0])}
        out = TensionResult(dictresult)
        return out

    return tension


def sgt_spot(rho1, rho2, Tsat, Psat, model, n=50, full_output=False,
             check_eq=True):
    """
    SGT spot for mixtures (rho1, rho2, T, P) -> interfacial tension

    Parameters
    ----------
    rho1 : float
        phase 1 density vector
    rho2 : float
        phase 2 density vector
    Tsat : float
        saturation temperature
    Psat : float
        saturation pressure
    model : object
        created with an EoS
    n : int, optional
        number points to solve density profiles
    full_output : bool, optional
        wheter to outputs all calculation info
    check_eq : bool, optional
        whether to check if given density vectors are in phase equilibria

    Returns
    -------
    ten : float
        interfacial tension between the phases
    """

    # adimensionalizar variables
    Tfactor, Pfactor, rofactor, tenfactor, zfactor = model.sgt_adim(Tsat)
    Pad = Psat*Pfactor
    ro1a = rho1*rofactor
    ro2a = rho2*rofactor

    cij = model.ci(Tsat)
    cij /= cij[0, 0]

    mu0 = model.muad(ro1a, Tsat)
    mu02 = model.muad(ro2a, Tsat)
    if check_eq:
        if not np.allclose(mu0, mu02, rtol=1e-3):
            raise Exception('Not equilibria compositions, mu1 != mu2')

    roots, weights = lobatto(n)
    s = 0
    try:
        ros = (ro1a + ro2a)/2
        ros = root(fobj_saddle, ros, args=(mu0, Tsat, model), method='lm')
        if ros.success:
            ros = ros.x
            # segment1
            # Linear profile
            pend = (ros - ro1a)
            b = ro1a
            ro1 = (np.outer(roots, pend) + b).T

            ro_s1 = (ros[s]-ro1a[s])*roots + ro1a[s]  # integration nodes
            wreal1 = np.abs(weights*(ros[s]-ro1a[s]))  # integration weights

            dro1 = np.gradient(ro1, ro_s1, edge_order=2, axis=1)

            # segment2
            # Linear profile
            pend = (ro2a - ros)
            b = ros
            ro2 = (np.outer(roots, pend) + b).T

            ro_s2 = (ro2a[s]-ro1a[s])*roots + ro1a[s]  # integration nodes
            wreal2 = np.abs(weights*(ro2a[s]-ro1a[s]))  # integration weights

            dro2 = np.gradient(ro2, ro_s2, edge_order=2, axis=1)

            dom1 = np.zeros(n)
            dom2 = np.zeros(n)
            for i in range(n):
                dom1[i] = model.dOm(ro1[:, i], Tsat, mu0, Pad)
                dom2[i] = model.dOm(ro2[:, i], Tsat, mu0, Pad)
            dom1[0] = 0.
            dom2[-1] = 0.

            suma1 = cmix_cy(dro1, cij)
            suma2 = cmix_cy(dro2, cij)

            integral1 = np.nan_to_num(np.sqrt(2*dom1*suma1))
            integral2 = np.nan_to_num(np.sqrt(2*dom2*suma2))
            tension = np.dot(integral1, wreal1)
            tension += np.dot(integral2, wreal2)
            tension *= tenfactor
            out = tension
            if full_output:
                with np.errstate(divide='ignore'):
                    intz1 = np.sqrt(suma1/(2*dom1))
                    intz2 = np.sqrt(suma2/(2*dom2))
                intz1[np.isinf(intz1)] = 0
                intz2[np.isinf(intz2)] = 0
                z1 = np.cumsum(intz1*wreal1)
                z2 = z1[-1] + np.cumsum(intz2*wreal2)
                z = np.hstack([z1, z2])
                ro = np.hstack([ro1, ro2])
                z /= zfactor
                ro /= rofactor
                dictresult = {'tension': tension, 'rho': ro, 'z': z,
                              'GPT': np.hstack([dom1, dom2])}
                out = TensionResult(dictresult)
        else:
            out = sgt_linear(ro1, ro2, Tsat, Psat, model, n, full_output)
    except:
        out = sgt_linear(ro1, ro2, Tsat, Psat, model, n, full_output)

    return out
